"""Module containing helper functions related to querying files."""

import json, os, re

from typing import Any, Union

from neurocaps.utils._logging import setup_logger

LG = setup_logger(__name__)


def setup_extraction(
    layout: Any,
    subj_ids: list[str],
    space: Union[str, None],
    exclude_niftis: Union[list[str], None],
    signal_clean_info: dict[str, Any],
    task_info: dict[str, Any],
    verbose: bool,
) -> tuple[list[str], dict[str, Any]]:
    """
    Get valid subjects (stored in ``self._subject_info``) and all required information needed
    to extract timeseries.

    Note
    ----
    The ``BIDSLayout`` type hint for ``layout`` is not added to allow certain in
    ``TimeseriesExtractor`` to be used on Windows machines that do not have pybids installed.
    """
    base_dict = {"layout": layout, "subj_id": None}
    subject_ids = []
    subject_info = {}

    for subj_id in subj_ids:
        base_dict["subj_id"] = subj_id
        files = build_dict(base_dict, space, signal_clean_info, task_info)

        # Remove excluded file from the niftis list, which will prevent it from being processed
        if exclude_niftis and files["niftis"]:
            files["niftis"] = exclude_nifti_files(files["niftis"], exclude_niftis)

        # Get subject header
        subject_header = create_header(subj_id, task_info)

        # Check files
        skip, msg = check_files(files, signal_clean_info, task_info)
        if msg and verbose:
            LG.warning(subject_header + msg)
        if skip:
            continue

        # Ensure only a single session is present if session is None
        if not task_info["session"]:
            check_sessions(files["niftis"], subject_header)

        # Generate a list of runs to iterate through based on runs in niftis
        check_runs = generate_runs_list(files["niftis"], task_info)
        if check_runs:
            run_list = filter_runs(check_runs, files, signal_clean_info, task_info)

            # Skip subject if no run has all needed files present
            if not run_list:
                if verbose:
                    LG.warning(
                        f"{subject_header}"
                        "Timeseries Extraction Skipped: None of the necessary files "
                        "(i.e NifTIs, confound tsv files, confound json files, event files) "
                        "are from the same run."
                    )
                continue

            if len(run_list) != len(check_runs):
                if verbose:
                    LG.warning(
                        f"{subject_header}"
                        "Only the following runs available contain all required files: "
                        f"{', '.join(run_list)}."
                    )
        elif not check_runs and task_info["runs"]:
            if verbose:
                requested_runs = [
                    f"run-{str(run)}" if "run-" not in str(run) else run
                    for run in task_info["runs"]
                ]
                LG.warning(
                    f"{subject_header}"
                    "Timeseries Extraction Skipped: Subject does not have any of the requested "
                    f"run IDs: {', '.join(requested_runs)}"
                )
                continue
        else:
            # Allows for nifti files that do not have the run- description
            run_list = [None]

        # Get repetition time for the subject
        tr = get_tr(files["bold_meta"], subject_header, signal_clean_info, task_info, verbose)

        # Add subject list to subject attribute. These are subjects that will be ran
        subject_ids.append(subj_id)

        # Store subject specific information
        subject_info[subj_id] = {"prepped_files": files, "tr": tr, "run_list": run_list}

    return subject_ids, subject_info


def query_files(
    layout: Any,
    task_info: dict[str, Any],
    extension: str,
    subj_id: str,
    scope: str = "derivatives",
    suffix: Union[str, None] = None,
    desc: Union[str, None] = None,
    event: bool = False,
    space: str = None,
):
    """
    Queries specific files (sorted lexicographically) using ``BidsLayout``.

    Note
    ----
    The type hint for ``layout`` is not added to allow certain in ``TimeseriesExtractor`` to be
    used on Windows machines that do not have pybids installed.
    """
    query_dict = {
        "scope": scope,
        "return_type": "file",
        "task": task_info["task"],
        "extension": extension,
        "subject": subj_id,
    }

    if desc:
        query_dict.update({"desc": desc})

    if suffix:
        query_dict.update({"suffix": suffix})

    if task_info["session"]:
        query_dict.update({"session": task_info["session"]})

    if not event and not desc:
        query_dict.update({"space": space})

    return sorted(layout.get(**query_dict))


def build_dict(
    base: dict[str, Union[Any, None]],
    space: str,
    signal_clean_info: dict[str, Any],
    task_info: dict[str, Any],
) -> dict[str, Union[str, None]]:
    """Builds dictionary containing subject-specific files queried using ``BIDSLayout``."""
    files = {}
    files["niftis"] = query_files(
        **base, task_info=task_info, suffix="bold", extension="nii.gz", space=space
    )

    files["bold_meta"] = query_files(
        **base, task_info=task_info, suffix="bold", extension="json", space=space
    )
    if not files["bold_meta"]:
        files["bold_meta"] = query_files(
            **base, scope="raw", task_info=task_info, suffix="bold", extension="json", space=None
        )

    if task_info["condition"]:
        files["events"] = query_files(
            **base, scope="raw", task_info=task_info, suffix="events", extension="tsv", event=True
        )
    else:
        files["events"] = []

    files["confounds"] = query_files(**base, task_info=task_info, desc="confounds", extension="tsv")

    if signal_clean_info["n_acompcor_separate"]:
        files["confound_metas"] = query_files(
            **base, task_info=task_info, desc="confounds", extension="json"
        )

    return files


def exclude_nifti_files(niftis: list[str], exclude_niftis: list[str]) -> list[str]:
    """Excludes certain NIfTI files based on ``exclude_niftis``."""
    exclude_niftis = exclude_niftis if isinstance(exclude_niftis, list) else [exclude_niftis]
    return [nifti for nifti in niftis if os.path.basename(nifti) not in exclude_niftis]


def create_header(subj_id: str, task_info: dict[str, Any]) -> str:
    """Creates base subject-specific header for logged messages."""
    sub_message = (
        f"[SUBJECT: {subj_id} | "
        f"SESSION: {task_info['session']} | "
        f"TASK: {task_info['task']}]"
    )
    subject_header = f"{sub_message} "

    return subject_header


def check_files(
    files: dict[str, list[str]], signal_clean_info: dict[str, Any], task_info: dict[str, Any]
) -> tuple[Union[bool, None], Union[str, None]]:
    """
    Simple initial check to ensure the required files are needed based on certain
    parameters ``__init__``.
    """
    skip, msg = None, None

    if not files["niftis"]:
        skip = True
        msg = (
            "Timeseries Extraction Skipped: No NIfTI files were found or all NifTI files "
            "were excluded."
        )

    if signal_clean_info["use_confounds"]:
        if not files["confounds"]:
            skip = True
            msg = (
                "Timeseries Extraction Skipped: `use_confounds` is True but no "
                "confound files found."
            )

        if signal_clean_info["n_acompcor_separate"] and not files.get("confound_metas"):
            skip = True
            msg = (
                "Timeseries Extraction Skipped: No confound metadata file found, which is "
                "needed to locate the first n components of the white-matter and cerebrospinal "
                "fluid masks separately."
            )

        if task_info["condition"] and not files["events"]:
            skip = True
            msg = (
                "Timeseries Extraction Skipped: `condition` is specified but no event files "
                "found."
            )

    return skip, msg


def check_sessions(niftis: list[str], subject_header: str) -> None:
    """
    Checks if all subject's NIfTI's are from a single session and returns an error if
    different sessions are detected.
    """
    ses_list = []

    for nifti in niftis:
        if "ses-" in os.path.basename(nifti):
            ses_list.append(re.search(r"ses-(\S+?)_", os.path.basename(nifti))[0][:-1])

    ses_list = sorted(set(ses_list))
    if len(ses_list) > 1:
        raise ValueError(
            f"{subject_header}"
            "`session` not specified but subject has more than one session: "
            f"{', '.join(ses_list)}. In order to continue timeseries extraction, the "
            "specific session to extract must be specified using `session`."
        )


def generate_runs_list(niftis: list[str], task_info: dict[str, Any]) -> set[str]:
    """
    Gets all the runs for a specific subject and filters if specific runs are requested.
    Returns set of run IDs sorted lexicographically.
    """
    check_runs = []

    for nifti in niftis:
        if "run-" in os.path.basename(niftis[0]):
            check_runs.append(re.search(r"run-(\S+?)_", os.path.basename(nifti))[0][:-1])

    check_runs = set(check_runs)

    requested_runs = {}
    if task_info["runs"]:
        requested_runs = {f"run-{run}" for run in task_info["runs"]}

    return sorted(check_runs.intersection(requested_runs)) if requested_runs else sorted(check_runs)


def filter_runs(
    check_runs: list[str],
    files: dict[str, list[str]],
    signal_clean_info: dict[str, Any],
    task_info: dict[str, Any],
) -> list[str]:
    """Filters runs by checking if all required files have the same run ID."""
    run_list = []

    # Check if at least one run has all files present
    for run in check_runs:
        bool_list = []

        # Assess is any of these returns True
        bool_list.append(any(f"{run}_" in file for file in files["niftis"]))

        if task_info["condition"]:
            bool_list.append(any(f"{run}_" in file for file in files["events"]))

        if signal_clean_info["use_confounds"]:
            bool_list.append(any(f"{run}_" in file for file in files["confounds"]))

            if signal_clean_info["n_acompcor_separate"]:
                bool_list.append(any(f"{run}_" in file for file in files["confound_metas"]))

        # Append runs that contain all needed files
        if all(bool_list):
            run_list.append(run)

    return run_list


def get_tr(
    bold_meta: str,
    subject_header: str,
    signal_clean_info: dict[str, Any],
    task_info: dict[str, Any],
    verbose: bool,
) -> Union[float, int, None]:
    """Gets repetition time."""
    try:
        if task_info["tr"]:
            tr = task_info["tr"]
        else:
            with open(bold_meta[0], "r") as json_file:
                tr = json.load(json_file)["RepetitionTime"]
    except (IndexError, KeyError, json.JSONDecodeError) as e:
        base_msg = (
            "`tr` not specified and could not be extracted since using the first BOLD "
            "metadata file"
        )
        base_msg += " due to " + (
            f"there being no BOLD metadata files for [TASK: {task_info['task']}]"
            if str(type(e).__name__) == "IndexError"
            else f"{type(e).__name__} - {str(e)}"
        )

        if task_info["condition"]:
            raise ValueError(
                f"{subject_header}"
                f"{base_msg}" + " The `tr` must be provided when `condition` is specified."
            )
        elif any(
            [
                signal_clean_info["masker_init"]["high_pass"],
                signal_clean_info["masker_init"]["low_pass"],
            ]
        ):
            raise ValueError(
                f"{subject_header}"
                f"{base_msg}"
                + " The `tr` must be provided when `high_pass` or `low_pass` is specified."
            )
        elif isinstance(signal_clean_info["fd_threshold"], dict) and signal_clean_info[
            "fd_threshold"
        ].get("interpolate"):
            raise ValueError(
                "`tr` must be provided when interpolation of censored volumes is required."
            )
        else:
            if verbose:
                LG.warning(
                    f"{subject_header}"
                    f"{base_msg}" + " `tr` has been set to None but extraction will continue."
                )
            tr = None

    return tr



"""Internal module for computing temporal dynamic metrics."""

import itertools, os
from typing import Union

import numpy as np
import pandas as pd
from numpy.typing import NDArray

from neurocaps.utils import _io as io_utils
from neurocaps.utils._helpers import list_to_str
from neurocaps.utils._logging import setup_logger

LG = setup_logger(__name__)


def filter_metrics(metrics: Union[list[str], tuple[str], None]) -> list[str]:
    """
    Filters metrics to ensure only the supported metrics ("temporal_fraction", "persistence",
    "counts", "transition_frequency") are in the list. Maintains the order of the original
    user-specified metrics.
    """
    metrics = (
        ("temporal_fraction", "persistence", "counts", "transition_frequency")
        if metrics is None
        else metrics
    )
    metrics = [metrics] if isinstance(metrics, str) else metrics
    metrics_set = set(metrics)

    valid_metrics = {
        "temporal_fraction",
        "persistence",
        "counts",
        "transition_frequency",
        "transition_probability",
    }
    set_diff = metrics_set - valid_metrics
    metrics_set = metrics_set.intersection(valid_metrics)
    # Ensure original order maintained
    ordered_metrics = [metric for metric in metrics if metric in metrics_set]

    if set_diff:
        LG.warning(f"The following invalid metrics will be ignored: {list_to_str(set_diff)}.")

    if not ordered_metrics:
        raise ValueError(
            f"No valid metrics in `metrics` list. Valid metrics are: {list_to_str(valid_metrics)}."
        )

    return ordered_metrics


def extract_caps_info(
    cap_dict: dict[str, dict[str, NDArray]],
) -> tuple[list[str], int, dict[str, int]]:
    """
    Extracts CAP-related information, specifically the names of the CAPs (i.e. CAP-1, CAP-2,
    etc), the maximum number of CAPs found across all groups (e.g if Group A has 4 CAPs and
    group B has 5 CAPs then 5 is returned), and the number of CAPs for each group.
    """
    group_cap_counts = {}

    for group_name in cap_dict:
        # Store the length of caps in each group
        group_cap_counts.update({group_name: len(cap_dict[group_name])})

    # CAP names based on groups with the most CAPs
    cap_names = list(cap_dict[max(group_cap_counts, key=group_cap_counts.get)])
    max_cap = max(group_cap_counts.values())

    return cap_names, max_cap, group_cap_counts


def create_transition_pairs(
    cap_dict: dict[str, dict[str, NDArray]],
) -> dict[str, list[tuple[int, int]]]:
    """Obtains all possible transition pairs."""
    group_caps = {
        group_name: [get_cap_id(cap_name) for cap_name in cap_dict[group_name]]
        for group_name in cap_dict
    }

    return {
        group_name: list(itertools.product(cap_id, cap_id))
        for group_name, cap_id in group_caps.items()
    }


def get_cap_id(cap_name: str) -> int:
    """
    Extracts the CAP name, assumed to be in the form `CAP-{n}` (i.e. CAP-1) and returns the int.
    """
    return int(cap_name.split("-")[-1])


def create_columns_names(
    metrics: list[str],
    group_names: list[str],
    cap_names: list[str],
    pairs: dict[str, list[tuple[int, int]]],
) -> dict[str, Union[list[str], dict[str, list[str]]]]:
    """
    Creates the column names for each requested metric. Used downstream has the column
    names for each metrics dataframe.
    """
    columns_names_dict = {}
    base_cols = ["Subject_ID", "Group", "Run"]

    for metric in metrics:
        if metric not in ["transition_frequency", "transition_probability"]:
            columns_names_dict.update({metric: base_cols + list(cap_names)})
        elif metric == "transition_probability":
            columns_names_dict[metric] = {}
            for group_name in group_names:
                col_names = base_cols + [f"{x}.{y}" for x, y in pairs[group_name]]
                columns_names_dict[metric].update({group_name: col_names})
        else:
            columns_names_dict.update({metric: base_cols + ["Transition_Frequency"]})

    return columns_names_dict


def initialize_all_metrics_dict(
    metrics: list[str], group_names: list[str]
) -> dict[str, Union[list, dict[str, list]]]:
    """
    Initializes a dictionary intended to store all computations for a metric across all
    subjects. The dictionary will be converted to a dataframe downstream.
    """
    all_metrics_dict = {}
    for metric in metrics:
        if metric != "transition_probability":
            all_metrics_dict[metric] = []
        else:
            all_metrics_dict["transition_probability"] = {
                group_name: [] for group_name in group_names
            }

    return all_metrics_dict


def create_distributed_dict(
    subject_table: dict[str, str], predicted_subject_timeseries: dict[str, NDArray]
) -> dict[str, list[tuple[str, str]]]:
    """Creates a dictionary mapping for each subject and run pair to iterate over."""
    distributed_dict = {}

    for subj_id, group_name in subject_table.items():
        distributed_dict[subj_id] = []
        for curr_run in predicted_subject_timeseries[subj_id]:
            distributed_dict[subj_id].append((group_name, curr_run))

    return distributed_dict


def compute_temporal_fraction(arr: NDArray, n_caps: int) -> dict[str, float]:
    """
    Computes temporal fraction for the subject and run specified in ``sub_info`` and inserts new
    row in the dataframe. Assumes one-based values in ``arr``.
    """
    frequency_dict = {key: np.where(arr == key, 1, 0).sum() for key in range(1, n_caps + 1)}
    proportion_dict = {key: value / (len(arr)) for key, value in frequency_dict.items()}

    return proportion_dict


def compute_counts(arr: NDArray, n_caps: int) -> dict[str, int]:
    """
    Computes counts for the subject and run specified in ``sub_info`` and inserts new row in the
    dataframe. Assumes one-based values in ``arr``.
    """
    count_dict = {}
    for target in range(1, n_caps + 1):
        if target in arr:
            _, counts = segments(target, arr)
            count_dict.update({target: counts})
        else:
            count_dict.update({target: 0})

    return count_dict


def compute_persistence(arr: NDArray, n_caps: int, tr: Union[float, int, None]) -> dict[str, float]:
    """
    Computes persistence for the subject and run specified in ``sub_info`` and inserts new row
    in the dataframe. Assumes one-based values in ``arr``.
    """
    persistence_dict = {}

    # Iterate through caps
    for target in range(1, n_caps + 1):
        # Floor n_segments at one to prevent nan due to division by 0
        binary_arr, n_segments = segments(arr, target, floor_at_one=True)
        persistence_dict.update({target: (binary_arr.sum() / n_segments) * (tr if tr else 1)})

    return persistence_dict


def segments(
    timeseries: NDArray, target: int, floor_at_one: bool = False
) -> tuple[NDArray[np.bool_], int]:
    """
    Computes the number of segments for persistence and counts computation. If ``floor_at_one``,
    then a minimum of 1 for number of segments is returned to prevent NaN due to divide by 0
    when computing persistence in which 0 is preferred to represent the absence of a CAP.

    Example Computation
    -------------------
    >>> import numpy as np
    >>> arr = np.array([1, 2, 1, 1, 1, 3])
    >>> target = 1
    >>> binary_arr = np.where(timeseries == target, 1, 0) # [1, 0, 1, 1, 1, 0]
    >>> target_indices = np.where(binary_arr == 1)[0] # [0, 2, 3, 4]
    >>> diff_arr = np.diff(target_indices, n=1) # [2, 1, 1]
    >>> n_transitions = np.where(diff_arr > 1, 1, 0).sum() # 1
    >>> n_segments += 1 # Account for first segment
    """
    binary_arr = np.where(timeseries == target, 1, 0)
    target_indices = np.where(binary_arr == 1)[0]
    n_segments = np.where(np.diff(target_indices, n=1) > 1, 1, 0).sum() + 1

    if not floor_at_one:
        n_segments = n_segments if binary_arr.sum() != 0 else 0

    return binary_arr, n_segments


def add_nans_to_dict(
    max_cap: int, n_group_caps: int, curr_dict: dict[str, Union[float, int]]
) -> dict[str, Union[float, int]]:
    """Adds NaN for groups with less CAPs than the group with the greatest number of CAPs."""
    if max_cap > n_group_caps:
        for i in range(n_group_caps + 1, max_cap + 1):
            curr_dict.update({i: float("nan")})

    return curr_dict


def convert_dict_to_df(
    columns_names_dict: dict[str, Union[list[str], dict[str, list[str]]]],
    all_metrics_dict: dict[str, Union[list, dict[str, list]]],
) -> dict[str, Union[pd.DataFrame, dict[str, pd.DataFrame]]]:
    """
    Appends the data ``all_metrics_dict`` to its respective dataframe in ``df_dict``.
    """
    df_dict = {}
    for metric_name in columns_names_dict:
        if metric_name != "transition_probability":
            df = pd.DataFrame.from_records(
                all_metrics_dict[metric_name], columns=columns_names_dict[metric_name]
            )
            df_dict[metric_name] = df
        else:
            df_dict["transition_probability"] = {}
            for group_name in columns_names_dict["transition_probability"]:
                df = pd.DataFrame.from_records(
                    all_metrics_dict[metric_name][group_name],
                    columns=columns_names_dict[metric_name][group_name],
                )
                df_dict[metric_name].update({group_name: df})

    return df_dict


def compute_transition_frequency(arr: NDArray) -> dict[str, int]:
    """
    Computes transition frequency for the subject and run specified in ``sub_info`` and inserts
    new row in the dataframe.

    Example Computation
    -------------------
    >>> import numpy as np
    >>> arr = np.array([1, 2, 1, 1, 1, 3])
    >>> n_trans = np.where(np.diff(arr, n=1) != 0, 1, 0).sum()
    """
    transition_frequency = np.where(np.diff(arr, n=1) != 0, 1, 0).sum()

    return {"transition_frequency": transition_frequency}


def compute_transition_probability(
    arr: NDArray, cap_pairs: list[tuple[int, int]]
) -> dict[str, float]:
    """
    Computes transition probability for the subject and run specified in ``sub_info`` and
    inserts new row in the dataframe.
    """
    trans_prob_dict = {}

    # Arrays for transitioning from and to element
    trans_from = arr[:-1]
    trans_to = arr[1:]

    # Iterate through pairs and calculate probability
    for e1, e2 in cap_pairs:
        # Get total number of possible transitions for first element
        total_trans = np.sum(trans_from == e1)
        # Compute sum of adjacent pairs of A -> B and divide
        trans_prob_dict[f"{e1}.{e2}"] = (
            np.sum((trans_from == e1) & (trans_to == e2)) / total_trans if total_trans > 0 else 0
        )

    return trans_prob_dict


def save_metrics(
    output_dir: str,
    group_names: list[str],
    df_dict: dict[str, Union[pd.DataFrame, dict[str, pd.DataFrame]]],
    prefix_filename: Union[str, None],
) -> None:
    """Saves the metric dataframes as csv files."""
    if not output_dir:
        return None

    for metric in df_dict:
        filename = io_utils.filename(base_name=f"{metric}", add_name=prefix_filename, pos="prefix")
        if metric != "transition_probability":
            df_dict[f"{metric}"].to_csv(
                path_or_buf=os.path.join(output_dir, f"{filename}.csv"), sep=",", index=False
            )
        else:
            for group_name in group_names:
                df_dict[f"{metric}"][group_name].to_csv(
                    path_or_buf=os.path.join(
                        output_dir, f"{filename}-{group_name.replace(' ', '_')}.csv"
                    ),
                    sep=",",
                    index=False,
                )



"""Internal module containing helper functions for ``CAP.caps2corr``."""

import numpy as np
from pandas import DataFrame
from scipy.stats import pearsonr, spearmanr

# Dictionary of scipy correlation functions
CORR_FUNC = {"pearson": pearsonr, "spearman": spearmanr}


def add_significance_values(df: DataFrame, corr_df: DataFrame, method: str, fmt: str) -> DataFrame:
    """Add p-values to each correlation value in the dataframe."""
    # Get p-values; use np.eye to make main diagonals equal zero; implementation of
    # tozCSS from:
    # https://stackoverflow.com/questions/25571882/pandas-columns-correlation-with-statistical-significance
    pval_df = df.corr(method=lambda x, y: CORR_FUNC[method](x, y)[1]) - np.eye(*corr_df.shape)
    # Add asterisk to values that meet the threshold
    pval_df = pval_df.map(
        lambda x: f"({format(x, fmt)})" + "".join(["*" for code in [0.05, 0.01, 0.001] if x < code])
    )
    # Add the p-values to the correlation matrix
    corr_df = corr_df.map(lambda x: f"{format(x, fmt)}") + " " + pval_df

    return corr_df



"""Internal module for projecting CAPs onto NifTI images."""

import inspect
from functools import lru_cache

import nibabel as nib, numpy as np
from nilearn import datasets, image
from scipy.spatial import KDTree


def cap_to_img(atlas_file, cap_vector, fwhm, knn_dict):
    """
    Projects cluster centroids (CAPs) on to the parcellation map. Also if specified, performs
    k-nearest neighbors and spatial smoothing on the NifTI image.

    Important
    ---------
    Assumes the first label, extracted from the parcellation, is the background label and skips it.
    """
    atlas = nib.load(atlas_file)
    atlas_fdata = atlas.get_fdata()
    # Create array of zeroes with same dimensions as atlas
    atlas_array = np.zeros_like(atlas_fdata)

    # Get array containing all labels in atlas to avoid issue if the first non-zero atlas label is
    # not 1
    target_array = np.unique(atlas_fdata)

    # Start at 1 to avoid assignment to the background label
    for indx, value in enumerate(cap_vector, start=1):
        atlas_array[atlas_fdata == target_array[indx]] = value

    stat_map = nib.Nifti1Image(atlas_array, atlas.affine, atlas.header)

    # Knn implementation to aid in coverage issues
    if knn_dict:
        stat_map = perform_knn(atlas, knn_dict, stat_map)

    # Add smoothing to stat map to help mitigate potential coverage issues
    if fwhm is not None:
        stat_map = image.smooth_img(stat_map, fwhm=fwhm)

    return stat_map


@lru_cache(maxsize=2)
def build_tree(atlas):
    """Uses scipy's ``KDTree`` to optimize k-nearest neighbors interpolation."""
    non_zero_indices = np.array(np.where(atlas.get_fdata() != 0)).T
    kdtree = KDTree(non_zero_indices)

    return kdtree, non_zero_indices


def get_remove_indices(atlas, remove_labels):
    """If requested, gets the coordinates containing a specified label to be removed."""
    remove_indxs = np.where(np.isin(atlas.get_fdata(), remove_labels))

    return remove_indxs


def perform_knn(atlas, knn_dict, stat_map):
    """
    Perform KNN to assist with coverage issues prior to plotting.

    .. important:: Modifies ``stat_map`` in place.
    """
    remove_labels = tuple(knn_dict["remove_labels"]) if knn_dict.get("remove_labels") else None

    # Get target indices
    target_indices = get_target_indices(
        atlas, knn_dict["reference_atlas"], knn_dict["resolution_mm"], remove_labels
    )

    # Get non-zero indices; Build kdtree for nearest neighbors
    kdtree, non_zero_indices = build_tree(atlas)

    # Get k
    k = knn_dict["k"]
    for target_indx in target_indices:
        # Get the nearest non-zero index
        distances, neighbor_indx = kdtree.query(target_indx, k=k)
        nearest_neighbors = non_zero_indices[neighbor_indx]

        if k == 1:
            nearest_neighbors = [nearest_neighbors]

        # Values of nearest neighbors
        neighbor_values = [
            stat_map.get_fdata()[tuple(nearest_neighbor)] for nearest_neighbor in nearest_neighbors
        ]

        new_value = (
            majority_vote(neighbor_values)
            if knn_dict["method"] == "majority_vote"
            else distance_weighted(neighbor_values, distances)
        )

        # Assign the new value to the current index
        stat_map.get_fdata()[tuple(target_indx)] = new_value

    return stat_map


@lru_cache(maxsize=4)
def get_target_indices(atlas, reference_atlas, resolution_mm, remove_labels):
    """
    Uses a reference atlas ("Schaefer" or "AAL") as a mask to identify non-background coordinates
    to use for k-nearest neighbors interpolation.
    """
    if reference_atlas == "Schaefer":
        reference_atlas_map = datasets.fetch_atlas_schaefer_2018(
            resolution_mm=resolution_mm, verbose=0
        )["maps"]
    else:
        reference_atlas_map = datasets.fetch_atlas_aal(verbose=0)["maps"]

    # Resample reference to atlas file using nearest interpolation to retain labels
    resampled_reference_atlas = resample_image(source_img=reference_atlas_map, target_img=atlas)

    # Get indices that equal zero in reference atlas to avoid interpolating background values
    reference_background_indices = set(zip(*np.where(resampled_reference_atlas.get_fdata() == 0)))

    # Get indices 0 indices for atlas
    zeroed_indices_atlas = set(zip(*np.where(atlas.get_fdata() == 0)))

    # Get the non-background indices through set subtraction
    if remove_labels:
        remove_indxs = get_remove_indices(atlas, remove_labels)
        remove_indxs = set(zip(*remove_indxs))
        target_indices = list(zeroed_indices_atlas - reference_background_indices - remove_indxs)
    else:
        target_indices = list(zeroed_indices_atlas - reference_background_indices)

    target_indices = sorted(target_indices)

    return target_indices


def resample_image(source_img, target_img):
    """Resamples source image to the target image."""
    kwargs = {
        "source_img": source_img,
        "target_img": target_img,
        "interpolation": "nearest",
        "force_resample": True,
    }

    if "copy_header" in inspect.signature(image.resample_to_img).parameters.keys():
        kwargs["copy_header"] = True

    return image.resample_to_img(**kwargs)


def distance_weighted(neighbor_values, distances):
    """Computes a new value based on the weighted inverse distance."""
    return np.mean(np.array(neighbor_values) * (1 / np.array(distances)))


def majority_vote(neighbor_values):
    """Selects new value based on highest frequency."""
    return max(set(neighbor_values), key=neighbor_values.count)



"""
Internal IO functions for creating directories, issuing warnings, serializing (with
Joblib or pickle), unserializing (with Joblib), and additional utilities.
"""

import os, copy, json, pickle
from typing import Any, Union

import joblib

from ._helpers import list_to_str
from ._logging import setup_logger
from neurocaps.exceptions import UnsupportedFileExtensionError

LG = setup_logger(__name__)


def makedir(output_dir: str) -> None:
    """Creates non-existent directory."""
    if output_dir and not os.path.exists(output_dir):
        LG.warning(f"Creating the following non-existent path: {output_dir}.")
        os.makedirs(output_dir)


def issue_file_warning(param_name: str, param: str, output_dir: str) -> None:
    """Issues warning."""
    if param is not None and not output_dir:
        LG.warning(
            f"`{param_name}` supplied but no `output_dir` specified. Files will not be saved."
        )


def filename(
    base_name: str, add_name: Union[str, None], pos: str, ext: Union[None, str] = None
) -> str:
    """
    Adds the file extension and the prefix or suffix to the file name depending on if ``pos``
    is "prefix" or "suffix".
    """
    if not add_name:
        filename = f"{base_name}.{ext}" if ext else base_name

        return filename.replace(" ", "_")

    if pos == "prefix":
        # Use - since prefix is currently for ``CAP.calculate_metrics``
        add_name = os.path.splitext(add_name.rstrip())[0].rstrip()
        filename = f"{add_name}-{base_name}"
    else:
        filename = f"{base_name}_{add_name.rstrip().replace(' ', '_')}"

    filename = filename.replace(" ", "_")

    return f"{filename}.{ext}" if ext else filename


def serialize(obj: Any, output_dir: str, save_filename: str, use_joblib=False) -> None:
    """
    Serialize an object with standard pickle or Joblib. Internally, plots use pickle for
    compatibility with pickle or Joblib.
    """
    with open(os.path.join(output_dir, save_filename), "wb") as f:
        joblib.dump(obj, f) if use_joblib else pickle.dump(obj, f)


def get_obj(obj: Any, needs_deepcopy=True) -> Any:
    """Determines if object needs deepcopy or needs to be unserialized."""
    if not isinstance(obj, str):
        obj = copy.deepcopy(obj) if needs_deepcopy else obj
        return obj
    else:
        return unserialize(obj)


def unserialize(filename: str, **kwargs) -> Any:
    """Opens serialized objects such as ``subject_timeseries`` or ``parcel_approach``."""
    base_ext = [".pkl", ".pickle", ".joblib"]
    supported_ext = [*base_ext, *[f"{x}.gz" for x in base_ext]]
    supported_ext += [".json"]
    ext = validate_file(filename, supported_ext, return_ext=True)
    if ext == ".json":
        obj = open_json_file(filename, **kwargs)
    else:
        obj = open_pickle_file(filename)

    return obj


def open_json_file(filename: str, **kwargs) -> dict:
    """Open a json file."""
    with open(filename, "r") as f:
        obj = json.load(f, **kwargs)

    return obj


def open_pickle_file(filename: str) -> dict:
    """Opens pickle file."""
    with open(filename, "rb") as f:
        obj = joblib.load(f)

    return obj


def check_file_exist(
    filename: str, raise_errors: bool = True, return_flag: bool = False
) -> Union[bool, None]:
    """Check if a file exists."""
    # Dont end with periods
    if not isinstance(filename, str) and raise_errors:
        raise ValueError(f"The following file must be a string: {filename}")

    file_exist = os.path.isfile(filename)
    if not file_exist and raise_errors:
        raise FileExistsError(f"The following file does not exist: {filename}")

    return file_exist if return_flag else None


def check_ext(
    filename: str, supported_ext: list[str], return_ext: bool = False
) -> Union[str, None]:
    """Checks if a file has a valid extension."""
    if filename.endswith(".gz"):
        _, ext = os.path.splitext(filename.removesuffix(".gz"))
        ext += ".gz"
    else:
        _, ext = os.path.splitext(filename)

    # Dont end with periods
    if ext not in supported_ext:
        raise UnsupportedFileExtensionError(
            f"The following file has an unsupported extension: {filename}\n"
            f"Only the following extensions are supported: {list_to_str(supported_ext)}"
        )

    return ext if return_ext else None


def validate_file(
    filename: str, supported_ext: list[str], return_ext: bool = False
) -> Union[str, None]:
    """Validates file and returns extension if True."""
    check_file_exist(filename)

    return check_ext(filename, supported_ext, return_ext)



"""Internal Function for checking confound names"""

from neurocaps.utils._logging import setup_logger

LG = setup_logger(__name__)


def check_confound_names(high_pass, user_confounds, n_acompcor_separate):
    """
    Pipeline for checking confound names. The default confound names ("basic") depend on whether
    high-pass filtering is specified.
    """
    if user_confounds == "basic":
        if high_pass:
            # Do not use cosine or acompcor regressor if high-pass filtering is not None.
            # Acompcor regressors are estimated on high pass filtered version of data form fmriprep
            confound_names = [
                "trans_x",
                "trans_x_derivative1",
                "trans_y",
                "trans_y_derivative1",
                "trans_z",
                "trans_z_derivative1",
                "rot_x",
                "rot_x_derivative1",
                "rot_y",
                "rot_y_derivative1",
                "rot_z",
                "rot_z_derivative1",
            ]
        else:
            confound_names = [
                "cosine*",
                "trans_x",
                "trans_x_derivative1",
                "trans_y",
                "trans_y_derivative1",
                "trans_z",
                "trans_z_derivative1",
                "rot_x",
                "rot_x_derivative1",
                "rot_y",
                "rot_y_derivative1",
                "rot_z",
                "rot_z_derivative1",
                "a_comp_cor_00",
                "a_comp_cor_01",
                "a_comp_cor_02",
                "a_comp_cor_03",
                "a_comp_cor_04",
                "a_comp_cor_05",
            ]
    else:
        assert (
            isinstance(user_confounds, list) and user_confounds
        ), "`confound_names` must be a non-empty list."
        confound_names = user_confounds

    confound_names = check_wildcard_confounds(confound_names)

    if n_acompcor_separate:
        confound_names = remove_a_comp_cor(confound_names, user_confounds, n_acompcor_separate)

    check_regressors(confound_names, n_acompcor_separate)

    LG.info(f"Confound regressors to be used if available: {', '.join(confound_names)}.")

    return confound_names.copy()


def remove_a_comp_cor(confound_names, user_confounds, n):
    """
    Removes all "a_comp_cor" regressors in ``confound_names`` if separate components for the white
    matter and cerebrospinal fluid masks are requested.
    """
    check_confounds = [confound for confound in confound_names if "a_comp_cor" not in confound]
    if len(confound_names) > len(check_confounds):
        removed_confounds = [
            element for element in confound_names if element not in check_confounds
        ]
        if user_confounds:
            LG.warning(
                "Since `n_acompcor_separate` has been specified, acompcor components in "
                f"`confound_names` will be disregarded and replaced with the first {n} "
                "components of the white matter and cerebrospinal fluid masks for each "
                "participant. The following components will not be used: "
                f"{', '.join(removed_confounds)}."
            )

    return check_confounds


def check_regressors(confound_names, n):
    """
    Performs a basic check to see if at least one "cosine" regressor is specified if "a_comp_cor"
    and "t_comp_cor" are detected in ``confound_names``.
    """
    cosine = any(i.startswith("cosine") for i in confound_names)
    acompcor = any(i.startswith("a_comp_cor") for i in confound_names) if not n else n
    tcompcor = any(i.startswith("t_comp_cor") for i in confound_names)
    if not cosine and (acompcor or tcompcor):
        LG.warning(
            "fMRIPrep applies high-pass filtering before running anatomical and temporal CompCor. "
            "It is recommended to include the discrete cosine-bases regressors ('cosine_XX') in "
            "`confound_names` if including any 'a_comp_cor' or 't_comp_cor' regressors."
        )


def check_wildcard_confounds(confound_names):
    """Checks asterisk position to ensure the wildcard expression is at the end of the word."""
    removed_confounds = []
    for confound_name in confound_names:
        if "*" in confound_name:
            indices = [indx for indx, element in enumerate(confound_name) if element == "*"]
            final_indx = len(confound_name) - 1
            if any(
                [len(indices) > 1, indices[0] != final_indx, not bool(confound_name.split("*")[0])]
            ):
                removed_confounds.append(confound_name)

    if removed_confounds:
        LG.warning(
            "Only wildcard prefixes (e.g. 'cosine*', 'rot*', etc) are supported. The following "
            f"confounds will be removed: {removed_confounds}."
        )

        confound_names = [
            confound_name
            for confound_name in confound_names
            if confound_name not in removed_confounds
        ]

    return confound_names



"""Module containing preset constants (kind of since some information is added at runtime)."""

# Consider adding to json if file gets too large
VALID_PRESETS = ["HCPex", "4S", "Gordon"]

PRESET_ATLAS_NAME = {
    "HCPex": "tpl-MNI152NLin2009cAsym_atlas-HCPex_2mm.nii.gz",
    "4S": "tpl-MNI152NLin2009cAsym_atlas-4S{}Parcels_res-01_dseg.nii.gz",
    "Gordon": "tpl-MNI_atlas-Gordon.nii.gz",
}

PRESET_JSON_NAME = {
    "HCPex": "atlas-HCPex_desc-CustomParcelApproach.json",
    "4S": "atlas-4S{}Parcels_desc-CustomParcelApproach.json",
    "Gordon": "atlas-Gordon_desc-CustomParcelApproach.json",
}

PRESET_METADATA = {
    "HCPex": {
        "name": "HCPex",
        "n_nodes": 426,
        "n_regions": 23,
        "space": "MNI152NLin2009cAsym",
        "doi": "10.1007/s00429-021-02421-6",
        "source": "https://github.com/wayalan/HCPex",
    },
    "4S": {
        "name": "4S",
        "n_nodes": None,
        "n_regions": 7,
        "space": "MNI152NLin2009cAsym",
        "source": "https://github.com/PennLINC/AtlasPack",
    },
    "Gordon": {
        "name": "Gordon",
        "n_nodes": 333,
        "n_regions": 13,
        "space": "MNI",
        "doi": "10.1093/cercor/bhu239",
        "source": "https://www.mir.wustl.edu/research/research-centers/neuroimaging/labs/egordon-lab/resources/",
    },
}

# TODO: Don't provide 856 for now, may not exist on OSF based on git annex
ATLAS_N_NODES = {
    "4S": {"valid_n": [156, 256, 356, 456, 556, 656, 756, 956, 1056], "default_n": 456}
}

OSF_FILE_URLS = {
    "tpl-MNI152NLin2009cAsym_atlas-HCPex_2mm.nii.gz": "mx4d6",
    "atlas-HCPex_desc-CustomParcelApproach.json": "rdbfv",
    "tpl-MNI152NLin2009cAsym_atlas-4S156Parcels_res-01_dseg.nii.gz": "fyd4e",
    "atlas-4S156Parcels_desc-CustomParcelApproach.json": "t3jkv",
    "tpl-MNI152NLin2009cAsym_atlas-4S256Parcels_res-01_dseg.nii.gz": "98tbx",
    "atlas-4S256Parcels_desc-CustomParcelApproach.json": "tuzk6",
    "tpl-MNI152NLin2009cAsym_atlas-4S356Parcels_res-01_dseg.nii.gz": "k9xcr",
    "atlas-4S356Parcels_desc-CustomParcelApproach.json": "2u3sa",
    "tpl-MNI152NLin2009cAsym_atlas-4S456Parcels_res-01_dseg.nii.gz": "tpz6y",
    "atlas-4S456Parcels_desc-CustomParcelApproach.json": "juyac",
    "tpl-MNI152NLin2009cAsym_atlas-4S556Parcels_res-01_dseg.nii.gz": "7d5xh",
    "atlas-4S556Parcels_desc-CustomParcelApproach.json": "m5xcg",
    "tpl-MNI152NLin2009cAsym_atlas-4S656Parcels_res-01_dseg.nii.gz": "2p4zt",
    "atlas-4S656Parcels_desc-CustomParcelApproach.json": "34qfc",
    "tpl-MNI152NLin2009cAsym_atlas-4S756Parcels_res-01_dseg.nii.gz": "6thzn",
    "atlas-4S756Parcels_desc-CustomParcelApproach.json": "hpxur",
    "tpl-MNI152NLin2009cAsym_atlas-4S956Parcels_res-01_dseg.nii.gz": "dw2bz",
    "atlas-4S956Parcels_desc-CustomParcelApproach.json": "9kmgb",
    "tpl-MNI152NLin2009cAsym_atlas-4S1056Parcels_res-01_dseg.nii.gz": "zsah9",
    "atlas-4S1056Parcels_desc-CustomParcelApproach.json": "a87c2",
    "tpl-MNI_atlas-Gordon.nii.gz": "ynpcu",
    "atlas-Gordon_desc-CustomParcelApproach.json": "dygfz",
}



"""Public utility functions for parcellations."""

from typing import Any, Literal, Optional, Union

import pandas as pd

from . import _io as io_utils
from ._helpers import list_to_str
from ._logging import setup_logger
from neurocaps.typing import CustomParcelApproach
from neurocaps.utils.datasets import _fetch as fetch_utils

LG = setup_logger(__name__)


def fetch_preset_parcel_approach(
    name: str, n_nodes: Optional[int] = None
) -> dict[Literal["Custom"], CustomParcelApproach]:
    """
    Fetches a Preset "Custom" Parcellation Approach.

    Creates a directory in the user home directory named "neurocaps_data" and downloads
    the ``parcel_approach`` (a JSON file) and the associated NifTI image from the
    Open Science Framework (OSF) if the corresponding files are not present in the
    directory.

    .. versionadded:: 0.32.2

    Parameters
    ----------
    name: :obj:`str`
        Name of the preset "Custom" parcellation approach to fetch. Options are "HCPex", "4S",
        and "Gordon".

        .. versionadded:: 0.32.3
           Added "Gordon".

    n_nodes: :obj:`int` or :obj:`None`, default=None
        Currently only relevant to "4S". Options for the "4S" are: 156, 256, 356,
        456, 556, 656, 757, 956, 1056. Defaults to 456 if None.

        .. note::
           The 856 node version of "4S" is currently unavailable.

    Returns
    -------
    dict[Literal["Custom"], CustomParcelApproach]
        A dictionary representing the "Custom" parcellation approach.

    Note
    ----
    **Region Mapping**: The mapping of regions/networks corresponds to the indices in the "nodes"
    list not its label ID. So, the first non-background index will be 0.

    See Also
    --------
    CustomParcelApproach
       The type definition for the Custom parcellation approach.
       (See `CustomParcelApproach Documentation
       <https://neurocaps.readthedocs.io/en/stable/api/generated/neurocaps.typing.CustomParcelApproach.html#neurocaps.typing.CustomParcelApproach>`_)

    ParcelApproach
        Type definition representing the structure of the Schaefer, AAL, and Custom parcellation
        approaches.
        (See `ParcelApproach Documentation
        <https://neurocaps.readthedocs.io/en/stable/api/generated/neurocaps.typing.ParcelApproach.html#neurocaps.typing.ParcelApproach>`_)
    """
    return fetch_utils.fetch_custom_parcel_approach(name, n_nodes)


def generate_custom_parcel_approach(
    filepath_or_df: Union[pd.DataFrame, str],
    maps_path: str,
    column_map: dict[Literal["nodes", "regions", "hemispheres"], str],
    hemisphere_map: Optional[dict[Literal["lh", "rh"], list[str]]] = None,
    background_label: Optional[str] = "Background",
    metadata: Optional[dict[str, Any]] = None,
) -> dict[Literal["Custom"], CustomParcelApproach]:
    """
    Generate a "Custom" Parcellation Approach From a Tabular Metadata File.

    Constructs the nested dictionary for the "Custom" ``parcel_approach`` by reading a tabular
    file containing metadata about the parcellation.

    .. important:
        The labels in the dataframe are assumed to be in order (minimum -> maximum label).

    .. versionadded:: 0.32.0

    Parameters
    ----------
    filepath_or_df: :obj:`pd.Dataframe` or :obj:`str`
        Path to the parcellation description file (e.g., "Schaefer_400.tsv"). Supports files
        with the following extension: ".csv", ".tsv", ".txt", ".xlsx". Can also be
        a pandas Dataframe.

    maps_path: :obj:`str`
        Path to the corresponding NIfTI parcellation file (e.g., "Schaefer_400.nii.gz").

    column_map: :obj:`dict[Literal["nodes", "regions", "hemispheres"], str]`
        A dictionary mapping keys to column names in the metadata file. The following keys are
        valid:

        - "nodes" (Required): Column name for the ROI labels/names.
        - "regions" (Optional): Column name for the network/region names.
        - "hemispheres" (Optional): Column name for hemisphere labels.

        .. note::
           If a "regions" key is provided in ``column_map``, then the "regions" subkey will be
           created. Additionally, for lateralization, a "hemispheres" key must be provided in
           ``column_map``. If the "regions" key is not provided, then the output custom
           parcellation dictionary will only contain the "maps" and "nodes" subkeys.

    hemisphere_map: :obj:`dict[Literal["lh", "rh"], list[str]]` or :obj:`None`, default=None
        A dictionary mapping hemisphere values in the metadata column to "lh" and "rh"
        (e.g. {"lh": ["Left", "L", "l", "LH"], "rh": ["Right", "R", "r", "RH"]]}).

        .. important::
           An error will be raised for cases where a region/network is partially lateralized
           (where not all nodes in the network are lateralized). Note that the lateralization
           information is only used in a specific case in ``CAP.caps2plot`` when
           ``visual_scope`` is set to "nodes" and the ``add_custom_node_labels`` kwarg is True.

    background_label: :obj:`str` or :obj:`None`, default="Background"
        The label name for the background ROI to be excluded (e.g. "background", "Unknown", etc).

        .. important::
           Will be used to only check the **first** row in the metadata file. Since it is
           assumed that the first label, typically the 0 index is the background. This label
           will be removed and the first non-background label is assigned as the 0 index in the
           "regions" sub-dictionary, which will shift the remaining labels backwards. Note that
           the mapping of regions/networks corresponds to the indices in the "nodes" list
           not its label ID.

    metadata: :obj:`dict[str, Any]` or None, default=None
        Metadata information to add to the "metadata" key if not None.

        .. versionadded:: 0.32.2

    Returns
    -------
    dict[Literal["Custom"], CustomParcelApproach]
        A dictionary representing the "Custom" parcellation approach.

    See Also
    --------
    CustomParcelApproach
       The type definition for the Custom parcellation approach.
       (See `CustomParcelApproach Documentation
       <https://neurocaps.readthedocs.io/en/stable/api/generated/neurocaps.typing.CustomParcelApproach.html#neurocaps.typing.CustomParcelApproach>`_)

    ParcelApproach
        Type definition representing the structure of the Schaefer, AAL, and Custom parcellation
        approaches.
        (See `ParcelApproach Documentation
        <https://neurocaps.readthedocs.io/en/stable/api/generated/neurocaps.typing.ParcelApproach.html#neurocaps.typing.ParcelApproach>`_)

    Note
    ----
    **Dataframe Structure**: The following is an example dataframe containing the metadata of
    the parcellation. The specific column names for the columns containing the labels (nodes),
    networks (regions), and hemisphere are to be mapped using ``column_map``. Note that the
    column names do not have to be named "labels", "networks", and "hemispheres". In this case,
    the map would be ``{"nodes": "labels", "regions": "networks", "hemispheres": "hemisphere}``.

    The labels for the hemispheres can be mapped using ``hemisphere_map``. In this case, the
    map would be ``{"lh": ["L"], "rh": [R, "rh"]}``.

    +------------------------+----------+------------+-----+
    | labels                 | networks | hemisphere | ... |
    +========================+==========+============+=====+
    | SalVentAttn_TempOccPar | VAN      | L          | ... |
    +------------------------+----------+------------+-----+
    | SalVentAttn_FrOperIns  | VAN      | R          | ... |
    +------------------------+----------+------------+-----+
    | Limbic_OFC             | Limbic   | rh         | ... |
    +------------------------+----------+------------+-----+
    | ...                    | ...      | ...        | ... |
    +------------------------+----------+------------+-----+

    **Excel Files**: Pandas requires an optional dependency (openpyxl) to be installed for
    Excel files (".xlsx").
    """
    _validate_column_map(column_map)
    _validate_hemisphere_map(hemisphere_map)

    df = _open_tabular_data(filepath_or_df)
    _check_if_columns_exists(df, column_map)

    # Drop first row if background label
    if background_label and df.loc[0, column_map["nodes"]] == background_label:
        df = df.iloc[1:].copy()

    # Check for NaN values
    for key in column_map:
        if key in ["nodes", "regions"]:
            _check_column_validity(df[column_map[key]])

    df = df.reset_index(drop=True)

    # Create base `parcel_approach`
    custom_parcel_approach = {"maps": str(maps_path), "nodes": df[column_map["nodes"]].tolist()}

    if "regions" in column_map:
        custom_parcel_approach["regions"] = _construct_regions_dict(df, column_map, hemisphere_map)

    if metadata:
        custom_parcel_approach["metadata"] = metadata

    return {"Custom": custom_parcel_approach}


def _validate_column_map(column_map: dict[Literal["nodes", "regions", "hemispheres"], str]) -> None:
    """Checks the validity of ``column_map``."""
    if not isinstance(column_map, dict):
        raise TypeError("`column_map` must be a dictionary.")

    required_keys = ["nodes"]
    optional_keys = ["regions", "hemispheres"]
    # Use for loop for now; may extend in future
    for key in required_keys:
        if key not in column_map:
            raise ValueError(f"`column_map` must contain a key for '{key}'.")

    # Check if value is string
    for key in column_map:
        if key not in required_keys + optional_keys:
            continue

        val = column_map.get(key)
        if not isinstance(val, str):
            raise ValueError(
                f"In `column_map`, '{key}' must be a string value but is currently assigned: {val}."
            )

    invalid_keys = set(column_map).difference(list(required_keys + optional_keys))
    if invalid_keys:
        LG.warning(
            f"The following keys are invalid and will be ignored: {list_to_str(invalid_keys)}."
        )


def _open_tabular_data(filepath_or_df: Union[pd.DataFrame, str]) -> pd.DataFrame:
    """Opens a tabular dataset."""
    if isinstance(filepath_or_df, pd.DataFrame):
        return filepath_or_df

    ext = io_utils.validate_file(filepath_or_df, [".csv", ".tsv", ".txt", ".xlsx"], return_ext=True)

    if ext != ".xlsx":
        df = pd.read_csv(filepath_or_df, sep=None, engine="python")
    else:
        df = pd.read_excel(filepath_or_df)

    return df


def _check_if_columns_exists(
    df: pd.DataFrame,
    column_map: dict[Literal["nodes", "regions", "hemispheres"], str],
) -> None:
    """Checks if the columns in ``column_map`` exists in the dataframe."""
    invalid_cols = set(column_map.values()).difference(df.columns)
    if invalid_cols:
        raise KeyError(
            "The following columns were not found in the metadata file: "
            f"{list_to_str(invalid_cols)}."
        )


def _validate_hemisphere_map(
    hemisphere_map: Optional[dict[Literal["lh", "rh"], list[str]]],
) -> None:
    """Validates ``hemisphere_map``."""
    if not hemisphere_map:
        return None

    if not isinstance(hemisphere_map, dict):
        raise TypeError("`hemisphere_map` must be a dictionary.")

    required_keys = ["lh", "rh"]
    missing_keys = set(required_keys).difference(hemisphere_map)
    if missing_keys:
        raise KeyError(
            "The following required keys are missing in ``hemisphere_dict``: "
            f"{list_to_str(missing_keys)}."
        )


def _construct_regions_dict(
    df: pd.DataFrame,
    column_map: dict[Literal["nodes", "regions", "hemispheres"], str],
    hemisphere_map: Optional[dict[Literal["lh", "rh"], list[str]]],
) -> dict[str, Union[dict[Literal["lh", "rh"], list[int]], list[int]]]:
    """Construct the "regions" dictionary."""
    regions_dict = {}

    # Only attempt to use hemispheres if the user mapped the column.
    attempt_lateralization = "hemispheres" in column_map

    if attempt_lateralization and hemisphere_map is None:
        # Use a default map if none is provided by the user.
        hemisphere_map = {"lh": ["L", "LH", "Left", "left"], "rh": ["R", "RH", "Right", "right"]}

    # Iterate through grouped dataframe -> (region_name, subset of data containing region_name)
    for region_name, group_df in df.groupby(column_map["regions"]):
        regions_dict[region_name] = group_df.index.tolist()

        if attempt_lateralization:
            if _is_group_lateralized(group_df, column_map["hemispheres"], hemisphere_map):
                lh_indices, rh_indices = _get_lateralized_region_indices(
                    group_df, column_map["hemispheres"], hemisphere_map
                )
                lateralized_indices = lh_indices + rh_indices
                _check_partial_lateralization(
                    group_df,
                    region_name,
                    column_map["nodes"],
                    column_map["hemispheres"],
                    lateralized_indices,
                )

                regions_dict[region_name] = {"lh": lh_indices, "rh": rh_indices}

    return regions_dict


def _has_nans_in_series(series: pd.Series) -> bool:
    """Checks for NaNs in a specific column."""
    return series.isna().sum()


def _check_column_validity(series: pd.Series) -> None:
    """Check if column as NaNs and is invalid (use for "nodes" and "regions")"""
    if _has_nans_in_series(series):
        raise ValueError(f"Must assign values to NaNs in the following column: {series.name}.")


def _is_group_lateralized(
    group_df: pd.DataFrame,
    hemisphere_col: str,
    hemisphere_map: dict[Literal["lh", "rh"], list[str]],
) -> bool:
    """Check if subset of data has any lateralized nodes."""
    return (
        group_df[hemisphere_col].notna().any()
        and group_df[hemisphere_col].isin(hemisphere_map["lh"] + hemisphere_map["rh"]).any()
    )


def _get_lateralized_region_indices(
    group_df: pd.DataFrame,
    hemisphere_col: str,
    hemisphere_map: dict[Literal["lh", "rh"], list[str]],
) -> tuple[list[int], list[int]]:
    """Get the indices for the left and right hemisphere from a grouped dataframe."""
    lh_indices = group_df[group_df[hemisphere_col].isin(hemisphere_map["lh"])].index.tolist()
    rh_indices = group_df[group_df[hemisphere_col].isin(hemisphere_map["rh"])].index.tolist()

    return lh_indices, rh_indices


def _check_partial_lateralization(
    group_df: pd.DataFrame,
    region_name: str,
    node_col: str,
    hemisphere_col: str,
    lateralized_indices: list[int],
) -> None:
    """
    Some regions are only partially lateralized (e.g. Cole-Anticevic atlas). This will result
    in an error.
    """
    if len(lateralized_indices) != len(group_df):
        unmapped_nodes = group_df[~group_df.index.isin(lateralized_indices)][node_col].tolist()
        raise ValueError(
            f"Region '{region_name}' has unmappable hemisphere labels for nodes: "
            f"{list_to_str(unmapped_nodes)}. Recommended to replace the hemisphere "
            "information of all nodes in this region to NaN in the following column: "
            f"{hemisphere_col}."
        )


__all__ = ["fetch_preset_parcel_approach", "generate_custom_parcel_approach"]



"""Internal module containing helper functions for ``CAP.get_caps``."""

from typing import Any, Union

import numpy as np
import matplotlib.pyplot as plt
from kneed import KneeLocator
from joblib import Parallel, delayed
from numpy.typing import NDArray
from matplotlib.figure import Figure
from sklearn.cluster import KMeans
from sklearn.metrics import davies_bouldin_score, calinski_harabasz_score, silhouette_score
from tqdm.auto import tqdm

from neurocaps.extraction._internals.postprocess import standardize_rois
from neurocaps.exceptions import NoElbowDetectedError
from neurocaps.typing import SubjectTimeseries
from neurocaps.utils import _io as io_utils
from neurocaps.utils._helpers import resolve_kwargs
from neurocaps.utils._logging import setup_logger
from neurocaps.utils._plotting_utils import PlotDefaults, PlotFuncs

LG = setup_logger(__name__)


def perform_kmeans(n_cluster: int, configs: dict, concatenated_timeseries: NDArray, method: str):
    """
    Uses scikit-learn to perform k-means clustering on concatenated timeseries data in both
    sequential and parallel contexts. Also uses scikit-learn to provide cluster performance metrics.
    """
    model = KMeans(n_clusters=n_cluster, **configs, verbose=0).fit(concatenated_timeseries)

    # Only return model when no cluster selection chosen
    if method is None:
        return model

    cluster_labels = model.labels_
    if method == "davies_bouldin":
        performance = {n_cluster: davies_bouldin_score(concatenated_timeseries, cluster_labels)}
    elif method == "elbow":
        performance = {n_cluster: model.inertia_}
    elif method == "silhouette":
        performance = {
            n_cluster: silhouette_score(concatenated_timeseries, cluster_labels, metric="euclidean")
        }
    else:
        # Variance Ratio
        performance = {n_cluster: calinski_harabasz_score(concatenated_timeseries, cluster_labels)}

    model_dict = {n_cluster: model}

    return performance, model_dict


def setup_groups(
    subject_timeseries: SubjectTimeseries, groups_dict: Union[dict[str, str], None]
) -> tuple[dict[str, str], dict[str, str]]:
    """Used to resolve ``self._groups`` and ``self._subject_table``."""
    if groups_dict is None:
        groups_dict = create_default_group(subject_timeseries)

    groups_dict = sort_subject_ids(groups_dict)

    subject_table = generate_lookup_table(groups_dict)

    return groups_dict, subject_table


def create_default_group(subject_timeseries: SubjectTimeseries):
    """
    Creates a dictionary mapping the default group "All Subjects" to the subject IDs in the
    SubjectTimeseries.
    """
    # Create dictionary for "All Subjects" if no groups are specified
    LG.info(
        "No groups specified. Using default group 'All Subjects' containing all subject "
        "IDs from `subject_timeseries`. The `groups` dictionary will remain fixed "
        "unless the `CAP` class is re-initialized or `clear_groups()` is used."
    )

    return {"All Subjects": list(subject_timeseries)}


def sort_subject_ids(group_dict: dict[str, str]) -> dict[str, str]:
    """Sort IDs lexicographically (also done in ``TimeseriesExtractor``)."""
    return {group: sorted(group_dict[group]) for group in group_dict}


def generate_lookup_table(group_dict: dict[str, str]) -> dict[str, str]:
    """Creates dictionary mapping subject IDs to their associated group."""
    subject_table = {}

    for group in group_dict:
        for subj_id in group_dict[group]:
            if subj_id in subject_table:
                LG.warning(
                    f"[SUBJECT: {subj_id}] Appears more than once. Only the first instance of "
                    "this subject will be included in the analysis."
                )
            else:
                subject_table.update({subj_id: group})

    return subject_table


def create_group_map(subject_table: dict[str, str], group_dict: dict[str, str]) -> dict[str, str]:
    """
    Create a new group_dict by intersecting ``subject_table`` and ``group_dict``.
    Done since ``subject_table`` will contain unique subject IDs but ``group_dict`` may have
    subject IDs that repeat across groups. Output used for tqdm progress bars.
    """
    # Intersect subjects in subjects table and the groups for tqdm
    group_map = {
        group_name: sorted(set(subject_table).intersection(group_dict[group_name]))
        for group_name in group_dict
    }

    return group_map


def concatenate_timeseries(
    subject_timeseries: SubjectTimeseries,
    group_dict: dict[str, str],
    runs: Union[list[int], list[str], None],
    progress_bar: bool,
) -> dict[str, NDArray]:
    """
    Concatenates the timeseries data of all subjects into a single numpy array if ``groups`` are
    None or group-specific numpy arrays.
    """
    # Collect timeseries data in lists
    group_arrays = {group_name: [] for group_name in group_dict}
    for group_name in group_dict:
        for subj_id in tqdm(
            group_dict[group_name],
            desc=f"Collecting Subject Timeseries Data [GROUP: {group_name}]",
            disable=not progress_bar,
        ):
            subject_runs, miss_runs = get_runs(runs, list(subject_timeseries[subj_id]))

            if miss_runs:
                LG.warning(
                    f"[SUBJECT: {subj_id}] Does not have the requested runs: "
                    f"{', '.join(miss_runs)}."
                )

            if not subject_runs:
                LG.warning(
                    f"[SUBJECT: {subj_id}] Excluded from the concatenated timeseries due to "
                    "having no runs."
                )
                continue

            subj_arrays = [subject_timeseries[subj_id][run_id] for run_id in subject_runs]
            group_arrays[group_name].extend(subj_arrays)

    # Only stack once per group; avoid bottleneck due to repeated calls on large data
    concatenated_timeseries = {group_name: None for group_name in group_dict}
    for group_name in tqdm(
        group_arrays, desc="Concatenating Timeseries Data Per Group", disable=not progress_bar
    ):
        concatenated_timeseries[group_name] = np.vstack(group_arrays[group_name])

    del group_arrays

    return concatenated_timeseries


def scale(concatenated_timeseries: dict[str, NDArray]) -> dict[str, NDArray]:
    """Scales the concatenated timeseries data."""
    mean_vec = {group_name: None for group_name in concatenated_timeseries}
    stdev_vec = {group_name: None for group_name in concatenated_timeseries}

    for group_name in concatenated_timeseries:
        concatenated_timeseries[group_name], mean_vec[group_name], stdev_vec[group_name] = (
            standardize_rois(concatenated_timeseries[group_name], return_parameters=True)
        )

    return concatenated_timeseries, mean_vec, stdev_vec


def get_runs(
    requested_runs: Union[list[int], list[str], None], curr_runs: list[str]
) -> tuple[list[str], Union[list[str], None]]:
    """
    Filters the current runs available for a subject if specific runs are requested.
    Also returns a list of missing runs that were requested
    """
    if requested_runs:
        requested_runs = [str(run).removeprefix("run-") for run in requested_runs]
        requested_runs = [f"run-{run}" for run in requested_runs]

    runs = [run for run in requested_runs if run in curr_runs] if requested_runs else curr_runs
    miss_runs = list(set(requested_runs) - set(runs)) if requested_runs else None

    return runs, miss_runs


def select_optimal_clusters(
    concatenated_timeseries_dict,
    method: str,
    n_clusters: list[int],
    n_cores: Union[int, None],
    configs: dict[str, Any],
    show_figs: bool,
    output_dir: Union[str, None],
    progress_bar: bool,
    as_pickle: bool,
    **kwargs,
) -> None:
    """Selects optimal number of clusters based on the specific ``cluster_selection_method``."""
    cluster_scores = {}
    optimal_n_clusters = {}
    kmeans = {}
    performance_dict = {}

    for group_name in concatenated_timeseries_dict:
        performance_dict[group_name] = {}
        model_dict = {}

        if n_cores is None:
            for n_cluster in tqdm(
                n_clusters, desc=f"Clustering [GROUP: {group_name}]", disable=not progress_bar
            ):
                output_score, model = perform_kmeans(
                    n_cluster, configs, concatenated_timeseries_dict[group_name], method
                )
                performance_dict[group_name].update(output_score)
                model_dict.update(model)
        else:
            parallel = Parallel(
                return_as="generator",
                n_jobs=n_cores,
                backend="loky",
                max_nbytes=kwargs.get("max_nbytes", "1M"),
            )
            outputs = tqdm(
                parallel(
                    delayed(perform_kmeans)(
                        n_cluster, configs, concatenated_timeseries_dict[group_name], method
                    )
                    for n_cluster in n_clusters
                ),
                desc=f"Clustering [GROUP: {group_name}]",
                total=len(n_clusters),
                disable=not progress_bar,
            )

            output_scores, models = zip(*outputs)
            for output in output_scores:
                performance_dict[group_name].update(output)
            for model in models:
                model_dict.update(model)

        # Select optimal clusters
        if method == "elbow":
            kneedle = KneeLocator(
                x=list(performance_dict[group_name]),
                y=list(performance_dict[group_name].values()),
                curve="convex",
                direction="decreasing",
                S=kwargs.get("S", 1.0),
            )

            optimal_n_clusters[group_name] = kneedle.elbow

            if optimal_n_clusters[group_name] is None:
                raise NoElbowDetectedError(
                    f"[GROUP: {group_name}] - No elbow detected. Try adjusting the sensitivity "
                    "parameter (`S`) to increase or decrease sensitivity (higher values "
                    "are less sensitive), expanding the list of `n_clusters` to test, or "
                    "using another `cluster_selection_method`."
                )
        elif method == "davies_bouldin":
            # Get minimum for davies bouldin
            optimal_n_clusters[group_name] = min(
                performance_dict[group_name], key=performance_dict[group_name].get
            )
        else:
            # Get max for silhouette and variance ratio
            optimal_n_clusters[group_name] = max(
                performance_dict[group_name], key=performance_dict[group_name].get
            )

        # Get the optimal kmeans model
        kmeans[group_name] = model_dict[optimal_n_clusters[group_name]]

        LG.info(
            f"[GROUP: {group_name} | METHOD: {method}] Optimal cluster size is "
            f"{optimal_n_clusters[group_name]}."
        )

        if show_figs or output_dir is not None:
            # Create plot dictionary
            filtered_kwargs = {k: v for k, v in kwargs.items() if k not in ["S", "max_nbytes"]}
            plot_dict = resolve_kwargs(PlotDefaults.get_caps(), **filtered_kwargs)

            fig = plot_cluster_performance(
                method,
                group_name,
                performance_dict[group_name],
                optimal_n_clusters[group_name] if method == "elbow" else None,
                show_figs,
                plot_dict,
            )

            save_cluster_performance_figure(
                fig, output_dir, group_name, method, as_pickle, plot_dict
            )

    cluster_scores = {"Cluster_Selection_Method": method}
    cluster_scores.update({"Scores": performance_dict})

    return optimal_n_clusters, kmeans, cluster_scores


def plot_cluster_performance(
    method: str,
    group_name: str,
    performance_dict: dict[str, float],
    optimal_n_clusters: Union[int, None],
    show_figs: bool,
    plot_dict: dict[str, Any],
) -> None:
    """Plots results of the specific ``cluster_selection_method``."""
    y_titles = {
        "elbow": "Inertia",
        "davies_bouldin": "Davies Bouldin Score",
        "silhouette": "Silhouette Score",
        "variance_ratio": "Variance Ratio Score",
    }

    plt.figure(figsize=plot_dict["figsize"])

    x_values = list(performance_dict)
    y_values = [y for _, y in performance_dict.items()]
    plt.plot(x_values, y_values)

    if plot_dict["step"]:
        x_ticks = range(x_values[0], x_values[-1] + 1, plot_dict["step"])
        plt.xticks(x_ticks)

    plt.title(group_name)
    plt.xlabel("K")

    y_title = y_titles[method]
    plt.ylabel(y_title)
    # Add vertical line for elbow method
    if y_title == "Inertia":
        plt.vlines(
            optimal_n_clusters,
            plt.ylim()[0],
            plt.ylim()[1],
            linestyles="--",
            label="elbow",
        )

    fig = plt.gcf()

    PlotFuncs.show(show_figs)

    return fig


def save_cluster_performance_figure(
    fig: Figure,
    output_dir: Union[str, None],
    group_name: str,
    method_name: str,
    as_pickle: bool,
    plot_dict: dict[str, Any],
) -> None:
    """Saves the cluster performance plot if ``output_dir`` is not falsy."""
    if not output_dir:
        return None

    io_utils.makedir(output_dir)

    save_name = f"{group_name.replace(' ', '_')}_{method_name}.png"
    PlotFuncs.save_fig(fig, output_dir, save_name, plot_dict, as_pickle)


def compute_variance_explained(
    concatenated_timeseries_dict: dict[str, NDArray], kmeans: dict[str, KMeans]
) -> dict[str, float]:
    """Computes variance explained in the concatenated timeseries by clustering."""
    variance_explained_dict = {}

    for group_name in concatenated_timeseries_dict:
        mean_vec = np.mean(concatenated_timeseries_dict[group_name], axis=0)
        total_var = np.sum((concatenated_timeseries_dict[group_name] - mean_vec) ** 2)
        explained_var = 1 - (kmeans[group_name].inertia_ / total_var)
        variance_explained_dict[group_name] = explained_var

    return variance_explained_dict


def create_caps_dict(kmeans_dict: dict[str, KMeans]) -> dict[str, NDArray]:
    """Maps groups to their CAPs (cluster centroids)."""
    caps_dict = {}

    for group_names in kmeans_dict:
        caps_dict[group_names] = {}
        cluster_centroids = zip(
            [num for num in range(1, len(kmeans_dict[group_names].cluster_centers_) + 1)],
            kmeans_dict[group_names].cluster_centers_,
        )
        caps_dict[group_names].update(
            {
                f"CAP-{state_number}": state_vector
                for state_number, state_vector in cluster_centroids
            }
        )

    return caps_dict



"""Contains the TimeseriesExtractor class for extracting timeseries"""

import os
from functools import lru_cache
from multiprocessing.queues import Queue
from typing import Any, Literal, Optional, Union
from typing_extensions import Self

from joblib import Parallel, delayed
from pandas import DataFrame
from tqdm.auto import tqdm

from ._internals import check_confound_names, process_subject_runs, TimeseriesExtractorGetter
from ._internals import bids_query, vizualization
from neurocaps.exceptions import BIDSQueryError
from neurocaps.typing import ParcelConfig, ParcelApproach, SubjectTimeseries
from neurocaps.utils import _io as io_utils
from neurocaps.utils._helpers import list_to_str, resolve_kwargs
from neurocaps.utils._logging import setup_logger
from neurocaps.utils._parcellation_validation import check_parcel_approach
from neurocaps.utils._plotting_utils import PlotDefaults, PlotFuncs

LG = setup_logger(__name__)


class TimeseriesExtractor(TimeseriesExtractorGetter):
    """
    Timeseries Extractor Class.

    Performs timeseries denoising, extraction, serialization (pickling), and visualization.

    .. important::
       It is highly recommended for all imaging data (the preprocessed BOLD images
       and parcellation map) to be in a standard MNI space. This is due to the ``CAP.caps2surf``
       assuming that the parcellation map is in MNI standard space when transforming data between
       volumetric and surface spaces.

    Parameters
    ----------
    space: :obj:`str`, default="MNI152NLin2009cAsym"
        The standard template space that the preprocessed BOLD data is registered to. This
        parameter is used for querying the preprocessed BOLD data with ``BIDSLayout``.

    parcel_approach: :obj:`ParcelConfig`, :obj:`ParcelApproach`, :obj:`str`, or :obj:`None`, default=None
        Specifies the parcellation approach to use. Options are "Schaefer", "AAL", or "Custom". Can
        be initialized with parameters, as a nested dictionary, or loaded from a serialized file
        (i.e. pickle, joblib, json). For detailed documentation on the expected structure, see the
        type definitions for ``ParcelConfig`` and ``ParcelApproach`` in the "See Also" section.
        Defaults to ``{"Schaefer": {"n_rois": 400, "yeo_networks": 7, "resolution_mm": 1}}`` if None.

        .. versionchanged:: 0.28.0
           Default changed to None; however, the same default dictionary configuration remains the
           same and is backwards compatible.

        .. versionchanged:: 0.31.0
           The default "regions" names for "AAL" has changed, which will group nodes differently.

    standardize: :obj:`bool` or or :obj:`None`, default=True
        Standardizes the timeseries (zero mean and unit variance using sample standard deviation).
        Always the final step in the pipeline.

        .. versionchanged:: 0.25.0
           No longer passed to Nilearn's ``NiftiLabelsMasker`` and only performs standardization
           using sample standard deviation. Default behavior of standardizing using sample standard
           deviation is the same; however, when not None or False, standardizing is always done at
           the end of the pipeline to prevent any external standardization from needing to be done
           when censoring or extracting condition.

        .. note::
           Standard deviations below ``np.finfo(std.dtype).eps`` are replaced with 1 for numerical
           stability.

    detrend: :obj:`bool`, default=False
        Detrends the timeseries.

        .. versionchanged:: 0.26.0
           Default changed from True to False due to the redundancy of detrending when discrete
           cosine-basis regressors are used, which is included in the "basic" option for
           ``confound_names``.

    low_pass: :obj:`float`, :obj:`int`, or :obj:`None`, default=None
        Filters out signals above the specified cutoff frequency.

    high_pass: :obj:`float`, :obj:`int`, or :obj:`None``, default=None
        Filters out signals below the specified cutoff frequency.

    fwhm: :obj:`float`, :obj:`int`, or :obj:`None`, default=None
        Applies spatial smoothing to data (in millimeters).

    use_confounds: :obj:`bool`, default=True
        If True, performs nuisance regression during timeseries extraction using the default or
        user-specified confounds in ``confound_names``.

        .. important::
           Requires the confound tsv files to be in same directory as preprocessed BOLD images.

    confound_names: {"basic"}, :obj:`list[str]`, or :obj:`None`, default="basic"
        Names of confounds extracted from the confound tsv files if ``use_confounds=True``.

        If "basic", the following confounds are used by default:

        - All discrete cosine-basis regressors.
        - Six head-motion parameters and their first-order derivatives.
        - First six principal aCompCor components.

        .. important::
            - Confound names follow fMRIPrep's naming scheme (versions >= 1.2.0).
            - Wildcards are supported to match prefixes (e.g., "cosine*" matches all confounds\
              starting with "cosine".)

    fd_threshold: :obj:`float`, :obj:`dict[str, float | int]`, or :obj:`None`, default=None
        Threshold for volume censoring based on framewise displacement (FD). Computed only after
        dummy volumes are removed.

        - *If float*, removes volumes where FD > threshold.
        - *If dict*, the following subkeys are available **(all non-required subkeys are None by default)**:

            - "threshold": A float (required). Removes volumes where FD > threshold.
            - "outlier_percentage": A float in interval [0,1]. Removes entire runs where proportion\
              of censored volumes exceeds this threshold. Proportion calculated after dummy scan removal.

            .. note::
                - A warning is issued when a run is flagged.
                - If ``condition`` specified for task-based data in ``self.get_bold()``, only\
                  considers volumes associated with the condition.

            - "n_before": An integer. Indicates the number of volumes to remove before each flagged volume.
            - "n_after": An integer. Indicates the number of volumes to remove after each flagged volume.
            - "use_sample_mask": A boolean. Controls when censoring is applied in the processing pipeline.

            .. important::
                - When True:

                    - Passes a sample mask of censored volumes to Nilearn's ``NiftiLabelsMasker``.
                    - Sets ``clean__extrapolate=False`` to prevent interpolation of end volumes.
                    - Censoring is applied before nuisance regression.
                    - If ``condition`` is specified for task-based data in ``self.get_bold()``, the\
                      timeseries is temporarily padded to extract the correct frames.

                - When False or None:

                    - Full timeseries data is used during nuisance regression.
                    - Censoring is applied after nuisance regression.

            - "interpolate": A boolean. If True, uses scipy's ``CubicSpline`` function` to perform\
              cubic spline interpolation only on censored frames. **Only performs interpolation if True**.

            .. note::
               Interpolation is only performed on frames that are bounded by non-censored
               frames on both ends. For example, given a
               ``censor_mask=[0, 0, 1, 0, 1, 0, 0, 0, 1, 0, 1, 0, 0]`` where "0" indicates censored
               high motion volumes and "1" indicates non-censored, low
               motion volumes, only the volumes at index 3, 5, 6, 7, and 9 would be interpolated.

        .. important::
            - A column named "framewise_displacement" must be available in the confounds file.
            - ``use_confounds`` must be set to True.
            - Do not specify "framewise_displacement" in ``confound_names``.
            - See Nilearn's documentation for details on censored volume handling when
              "use_sample_mask" is True:

                - `Signal Clean <https://nilearn.github.io/stable/modules/generated/nilearn.signal.clean.html>`_
                - `NiftiLabelsMasker <https://nilearn.github.io/stable/modules/generated/nilearn.maskers.NiftiLabelsMasker.html>`_

            - If "interpolate" is True, then interpolation is only applied after the nuisance
              regression and parcellation steps have been completed.
            - See Scipy's `CubicSpline <https://docs.scipy.org/doc/scipy/reference/generated/scipy.interpolate.CubicSpline.html>`_
              documentation.

    n_acompcor_separate: :obj:`int` or :obj:`None`, default=None
        Number of aCompCor components to extract separately from the white-matter (WM) and CSF
        masks. Uses first "n" components from each mask separately. For instance, if
        ``n_acompcor_separate=5``, then the the first 5 WM components and the first 5 CSF components
        (totaling 10 components) are regressed out.

        .. important::
            - ``use_confounds`` must be set to True.
            - If specified, this parameter overrides any aCompCor components listed in\
              ``confound_names``.

    dummy_scans: :obj:`int`, :obj:`dict[str, bool | int]`, "auto", or :obj:`None`, default=None
        Number of initial volumes to remove before timeseries extraction.

        - *If int*, removes first "n" volumes.
        - *If auto*, removes first "n" volumes based on "non_steady_state_outlier_XX" columns.
        - *If dict*, the following keys are supported **(all non-required subkeys are None by default)**:

            - "auto": A boolean (required). Removes first "n" volumes based on\
              "non_steady_state_outlier_XX" columns.
            - "min": An integer. Minimum volumes to remove when auto is set to True. If "auto"\
              finds 2 outliers but ``{"min": 3}``, removes 3 volumes.
            - "max": An integer. Maximum volumes to remove when auto is set to True. If "auto"\
              finds 6 outliers but ``{"max": 5}``, removes 5 volumes.

        .. important::
            - "auto" and dictionary option requires ``use_confounds`` to be True and\
              "non_steady_state_outlier_XX" to be present in the confounds tsv file.
            - "min" and "max" keys only apply when "auto" is True.

    dtype: :obj:`str` or "auto", default=None
        The NumPy dtype to convert NIfTI images to.


    Properties
    ----------
    space: :obj:`str`
        The standard template space that the preprocessed BOLD data is registered to. This property
        is also settable.

    parcel_approach: :obj:`ParcelApproach`
        Parcellation information with "maps" (path to parcellation file), "nodes" (labels), and
        "regions" (anatomical regions or networks). Returns a deep copy.

    signal_clean_info: :obj:`dict[str, bool | int | float | str]` or :obj:`None`
        Dictionary containing signal cleaning parameters. Returns a deep copy.

    task_info: :obj:`dict[str, str | int]` or :obj:`None`
        Dictionary containing all task-related information such. Defined after running
        ``self.get_bold()``.

    subject_ids: :obj:`list[str]` or :obj:`None`
        A list containing all subject IDs retrieved from ``BIDSLayout`` for timeseries extraction.
        Defined after running ``self.get_bold()``.

    n_cores: :obj:`int` or :obj:`None`
        Number of cores used for multiprocessing with Joblib. Defined after running
        ``self.get_bold()``.

    subject_timeseries: :obj:`SubjectTimeseries` or :obj:`None`
        A dictionary mapping subject IDs to their run IDs and their associated timeseries
        (TRs x ROIs) as a NumPy array. Can be deleted using ``del self.subject_timeseries``.
        Defined after running ``self.get_bold()``. This property is also settable (accepts a
        dictionary or pickle file). Returns a reference.

    qc: :obj:`dict[str, dict[str, dict[str, float | int]]]` or :obj:`None`
        A dictionary reporting quality control, which maps subject IDs to their run IDs and
        information related to framewise displacement and dummy scans. Returns a reference.

        ::

            {"subjectID": {"run-ID": {"mean_fd": float, "std_fd": float, ...}}}

    See Also
    --------
    :class:`neurocaps.typing.ParcelConfig`
        Type definition representing the configuration options and structure for the Schaefer
        and AAL parcellations.
        (See `ParcelConfig Documentation
        <https://neurocaps.readthedocs.io/en/stable/api/generated/neurocaps.typing.ParcelConfig.html#neurocaps.typing.ParcelConfig>`_)

    :class:`neurocaps.typing.ParcelApproach`
        Type definition representing the structure of the Schaefer, AAL, and Custom parcellation
        approaches.
        (See `ParcelApproach Documentation
        <https://neurocaps.readthedocs.io/en/stable/api/generated/neurocaps.typing.ParcelApproach.html#neurocaps.typing.ParcelApproach>`_)

    :data:`neurocaps.typing.SubjectTimeseries`
        Type definition representing the structure of the subject timeseries.
        (See: `SubjectTimeseries Documentation
        <https://neurocaps.readthedocs.io/en/stable/api/generated/neurocaps.typing.SubjectTimeseries.html#neurocaps.typing.SubjectTimeseries>`_)

    Note
    ----
    **Passed Parameters**: ``detrend``, ``low_pass``, ``high_pass``, ``fwhm``, and nuisance
    regressors (``confound_names``) uses ``nilearn.maskers.NiftiLabelsMasker``. The ``dtype``
    parameter is used by ``nilearn.image.load_img``.

    **Custom Parcellations**: Refer to the `NeuroCAPs' Parcellation Documentation
    <https://neurocaps.readthedocs.io/en/stable/parcellations.html>`_ for detailed explanations and
    example structures for Custom parcellations.
    """

    def __init__(
        self,
        space: str = "MNI152NLin2009cAsym",
        parcel_approach: Union[ParcelConfig, ParcelApproach, str, None] = None,
        standardize: bool = True,
        detrend: bool = False,
        low_pass: Optional[Union[float, int]] = None,
        high_pass: Optional[Union[float, int]] = None,
        fwhm: Optional[Union[float, int]] = None,
        use_confounds: bool = True,
        confound_names: Optional[Union[list[str], Literal["basic"]]] = "basic",
        fd_threshold: Optional[Union[float, dict[str, Union[bool, float, int]]]] = None,
        n_acompcor_separate: Optional[int] = None,
        dummy_scans: Optional[Union[int, dict[str, Union[bool, int]], Literal["auto"]]] = None,
        dtype: Optional[Union[str, Literal["auto"]]] = None,
    ) -> None:

        self._space = space
        # Check parcel_approach
        self._parcel_approach = check_parcel_approach(parcel_approach=parcel_approach)

        if use_confounds:
            if confound_names:
                # Replace confounds if not None
                confound_names = check_confound_names(
                    high_pass, confound_names, n_acompcor_separate
                )

            if fd_threshold:
                self._validate_init_params("fd_threshold", fd_threshold)

            if dummy_scans:
                self._validate_init_params("dummy_scans", dummy_scans)
        else:
            if confound_names:
                LG.warning(
                    "`confound_names` is specified but `use_confounds` is not True so nuisance "
                    "regression will not be done."
                )
                confound_names = None

            if fd_threshold:
                raise ValueError(
                    "`fd_threshold` specified but `use_confounds` is not True, so removal of "
                    "volumes after nuisance regression cannot be done since confounds tsv file "
                    "generated by fMRIPrep is needed."
                )

            if dummy_scans == "auto" or (isinstance(dummy_scans, dict) and dummy_scans.get("auto")):
                raise ValueError(
                    "'auto' specified in `dummy_scans` but `use_confounds` is not True, so "
                    "automated dummy scans detection cannot be done since confounds tsv file "
                    "generated by fMRIPrep is needed."
                )

            if n_acompcor_separate:
                raise ValueError(
                    "`n_acompcor_separate` specified and `use_confounds` is not True, so separate "
                    "WM and CSF components cannot be regressed out since confounds tsv file "
                    "generated by fMRIPrep is needed."
                )

        self._signal_clean_info = {
            "masker_init": {
                "detrend": detrend,
                "low_pass": low_pass,
                "high_pass": high_pass,
                "smoothing_fwhm": fwhm,
            },
            "standardize": standardize,
            "use_confounds": use_confounds,
            "confound_names": confound_names,
            "n_acompcor_separate": n_acompcor_separate,
            "dummy_scans": dummy_scans.copy() if isinstance(dummy_scans, dict) else dummy_scans,
            "fd_threshold": fd_threshold.copy() if isinstance(fd_threshold, dict) else fd_threshold,
            "dtype": dtype,
        }

    @staticmethod
    def _validate_init_params(param: str, struct: Any) -> None:
        """Validates structure for ``dummy_scans`` and ``fd_threshold``."""
        mandatory_keys = {
            "dummy_scans": {"auto": (bool,)},
            "fd_threshold": {"threshold": (float, int)},
        }

        optional_keys = {
            "dummy_scans": {"min": (int,), "max": (int,)},
            "fd_threshold": {
                "n_before": (int,),
                "n_after": (int,),
                "outlier_percentage": (float,),
                "use_sample_mask": (bool,),
                "interpolate": (bool,),
            },
        }

        type2str = lambda t: t.__name__
        str2text = lambda valid_types: ", ".join([type2str(t) for t in valid_types])

        valid_types = (dict, int, str) if param == "dummy_scans" else (dict, float, int)
        if not isinstance(struct, valid_types):
            raise TypeError(
                f"`{param}` must be one of the following types when not None: "
                f"{str2text(valid_types)}."
            )

        if param == "dummy_scans" and (isinstance(struct, str) and not struct == "auto"):
            raise ValueError(f"'auto' is the only valid string for `dummy_scans`.")

        if isinstance(struct, dict):
            # Check mandatory keys
            key = list(mandatory_keys[param].keys())[0]
            if key not in struct:
                raise KeyError(f"'{key}' is a mandatory key when `{param}` is a dictionary.")
            if not isinstance(struct[key], mandatory_keys[param][key]):
                raise TypeError(
                    f"'{key}' must be of one of the following types: "
                    f"{str2text(mandatory_keys[param][key])}."
                )

            # Check optional keys
            for key in optional_keys[param]:
                if key in struct:
                    if struct[key] is not None and not isinstance(
                        struct[key], optional_keys[param][key]
                    ):
                        raise TypeError(
                            f"'{key}' must be either None or of type "
                            f"{str2text(optional_keys[param][key])}."
                        )
                    # Additional check for "outlier_percentage"
                    if key == "outlier_percentage" and not 0 < struct["outlier_percentage"] < 1:
                        raise ValueError(
                            "'outlier_percentage' must be either None or a float between 0 and 1."
                        )

            # Check invalid keys
            set_diff = set(struct) - set(optional_keys[param]) - set(mandatory_keys[param])
            if set_diff:
                LG.warning(
                    f"The following invalid keys have been found in `{param}` and will be ignored: "
                    f"{list_to_str(set_diff)}."
                )

    def get_bold(
        self,
        bids_dir: str,
        task: str,
        session: Optional[Union[int, str]] = None,
        runs: Optional[Union[int, str, list[int], list[str]]] = None,
        condition: Optional[str] = None,
        condition_tr_shift: int = 0,
        tr: Optional[Union[int, float]] = None,
        slice_time_ref: Union[int, float] = 0.0,
        run_subjects: Optional[Union[str, list[str]]] = None,
        exclude_subjects: Optional[Union[str, list[str]]] = None,
        exclude_niftis: Optional[Union[str, list[str]]] = None,
        pipeline_name: Optional[str] = None,
        n_cores: Optional[int] = None,
        parallel_log_config: Optional[dict[str, Union[Queue, int]]] = None,
        verbose: bool = True,
        flush: bool = False,
        progress_bar: bool = False,
    ) -> Self:
        """
        Retrieve Preprocessed BOLD Data from BIDS Datasets.

        Extracts the timeseries data from preprocessed BOLD images located in the derivatives folder
        of a BIDS-compliant dataset. The timeseries data of all subjects are appended to a single
        dictionary ``self.subject_timeseries``.

        .. important::
            - For proper querying, a "dataset_description.json" file must be located in the root of
              the BIDs directory and the pipeline directory (located in the derivatives folder).
            - Refer to `NeuroCAPs' BIDS Structure and Entities documentation\
              <https://neurocaps.readthedocs.io/en/stable/bids.html>`_ for additional information
              about the expected directory structure and file naming scheme (entities) needed for
              querying.
            - This pipeline is most optimized for BOLD data preprocessed by fMRIPrep.

        Parameters
        ----------
        bids_dir: :obj:`str`
            Path to a BIDS compliant directory. A "dataset_description.json" file must be located
            in this directory or an error will be raised.

        task: :obj:`str`
            Name of task to extract timeseries data from (i.e "rest", "n-back", etc).

        session: :obj:`int`, :obj:`str`, or :obj:`None`, default=None
            The session ID to extract timeseries data from. Only a single session can be extracted
            at a time. The value can be an integer (e.g. ``session=2``) or a string (e.g.
            ``session="001"``).

        runs: :obj:`int`, :obj:`str`, :obj:`list[int]`, :obj:`list[str]`, or :obj:`None`, default=None
            List of run numbers to extract timeseries data from (e.g. ``runs=["000", "001"]``).
            Extracts all runs if unspecified.

        condition: :obj:`str` or :obj:`None`, default=None
            Isolates the timeseries data corresponding to a specific condition (listed in the
            "trial_type" column of the "events.tsv" file) after the timeseries has been extracted
            and subjected to nuisance regression. Only a single condition can be extracted at a time.

        condition_tr_shift: :obj:`int`, default=0
            Number of TR units to units to offset both the start and end scan indices of a condition
            to account for a fixed hemodynamic delay. This parameter only applies when a ``condition``
            is specified. For more details about how this offset affects the calculation of task
            conditions, see the "Extraction of Task Conditions" section below.

        tr: :obj:`int`, :obj:`float` or :obj:`None`, default=None
            Repetition time (TR), in seconds, for the specified task. If not provided, the TR will
            be automatically extracted from the first BOLD metadata file found for the task,
            searching first in the pipeline directory, then in the ``bids_dir`` if not found.

        slice_time_ref: :obj:`int` or :obj:`float`, default=0.0
            The reference slice expressed as a fraction of the ``tr`` that is subtracted from the
            condition onset times to adjust for slice time correction when ``condition`` is not None.
            Values can range from 0 to 1. For more details, see the "Extraction of Task Conditions"
            section below.

        run_subjects: :obj:`str`, :obj:`list[str]` or :obj:`None`, default=None
            A string (if single subject) or list of subject IDs to process (e.g.
            ``run_subjects=["01", "02"]``). Processes all subjects if None.

        exclude_subjects: :obj:`str`, :obj:`list[str]` or :obj:`None`, default=None
            A string (if single subject) or list of subject IDs to exclude (e.g.
            ``exclude_subjects=["01", "02"]``).

        exclude_niftis: :obj:`str`, :obj:`list[str]` or :obj:`None`, default=None
            A string (if single file) or List of the specific preprocessed NIfTI files to exclude,
            preventing their timeseries data from being extracted. Used if there are specific runs
            across different participants that need to be excluded.

        pipeline_name: :obj:`str` or :obj:`None`, default=None
            The name of the pipeline folder in the derivatives folder containing the preprocessed
            data. Used if multiple pipeline folders exist in the derivatives folder or the pipeline
            folder is nested (e.g. "fmriprep/fmriprep-20.0.0").

        n_cores: :obj:`int` or :obj:`None`, default=None
            The number of cores to use for multiprocessing with Joblib. The "loky" backend is used.

        parallel_log_config: :obj:`dict[str, multiprocessing.Manager.Queue | int]`
            Passes a user-defined managed queue and logging level to the internal timeseries
            extraction function when parallel processing (``n_cores``) is used. Available dictionary
            keys are:

            - "queue": The instance of ``multiprocessing.Manager.Queue`` to pass to ``QueueHandler``.
              If not specified, all logs will output to ``sys.stdout``.
            - "level": The logging level (e.g. ``logging.INFO``, ``logging.WARNING``). If not
              specified, the default level is ``logging.INFO``.

            Refer to the `NeuroCAPs' Logging documentation\
            <https://neurocaps.readthedocs.io/en/stable/logging.html>`_ for a detailed example
            setting up this parameter.

        verbose: :obj:`bool`, default=True
            If True, logs detailed subject-specific information including: subjects skipped due to
            missing required files, current subject being processed for timeseries extraction,
            confounds identified for nuisance regression in addition to requested confounds that are
            missing for a subject, and additional warnings encountered during the timeseries
            extraction process.

        flush: :obj:`bool`, default=False
            If True, flushes the logged subject-specific information produced during the timeseries
            extraction process.

        progress_bar: :obj:`bool`, default=False
            If True, displays a progress bar.

        Returns
        -------
        self

        Raises
        ------
        BIDSQueryError
            Occurs when subject IDs were not found during querying.

        See Also
        --------
        :data:`neurocaps.typing.SubjectTimeseries`
            Type definition representing the structure of the subject timeseries.

        Important
        ---------
        **Subject Timeseries Dictionary**: This function stores the extracted timeseries of all
        subjects in the ``subject_timeseries`` property and can be deleted using
        ``del self.subject_timeseries`` (Note that ``self.timeseries_to_pickle()`` and
        ``self.visualize_bold()`` need this property in order to be used). The structure is a
        dictionary mapping subject IDs to their run IDs and their associated timeseries (TRs x ROIs)
        as NumPy array. Refer to documentation for ``SubjectTimeseries`` in the "See Also" section
        for an example structure.

        **Data/Property Persistence**: Each time this function is called, it's associated properties
        such as ``self.subject_timeseries``, ``self.task_info``, ``self.qc``, etc, are automatically
        initialized/overwritten to create a clean state for the subsequent analysis. To save, the
        subject timeseries dictionary, ``self.timeseries_to_pickle()`` can be used. Additionally, to
        save the quality control dictionary, ``self.report_qc()`` can be used.

        **NIfTI Files Without "run-" Entity**: By default, "run-0" will be used as a placeholder,
        if run IDs are not specified in the NIfTI file.

        **Parcellation & Nuisance Regression**: For timeseries extraction, nuisance regression, and
        spatial dimensionality reduction using a parcellation, Nilearn's ``NiftiLabelsMasker``
        function is used. If requested, dummy scans are removed from the NIfTI images and confound
        dataset prior to timeseries extraction. For volumes exceeding a specified framewise
        displacement (FD) threshold, if the "use_sample_mask" key in the ``fd_threshold``
        dictionary is set to True, then a boolean sample mask is generated (where False indicates
        the high motion volumes) and passed to the ``sample_mask`` parameter in Nilearn's
        ``NiftiLabelsMasker``. If, "use_sample_mask" key is False or not specified in the
        ``fd_threshold`` dictionary, then censoring is done after nuisance regression, which is the
        default behavior.

        **Extraction of Task Conditions**: The formula used for computing the scan indices
        corresponding to the corresponding to a specific condition:

        ::

            adjusted_onset = condition_df.loc[i, "onset"] - slice_time_ref * tr
            onset_scan = math.floor(adjusted_onset / tr)
            onset_scan += condition_tr_shift
            end_scan = math.ceil((adjusted_onset + condition_df.loc[i, "duration"]) / tr)
            end_scan += condition_tr_shift

            onset_scan = max([0, onset_scan])
            end_scan = max([0, end_scan])
            scans.extend(range(onset_scan, end_scan))
            scans = sorted(list(set(scans)))

        .. versionchanged:: 0.28.4
           Max check done for ``onset_scan`` and ``end_scan`` instead of ``adjusted_onset``.

        When partial scans are computed, ``math.floor`` is used to round down for the beginning scan
        index and ``math.ceil`` is used to round up for the ending scan index. Negative scan indices
        are set to 0 to avoid unintentional negative indexing. For simplicity, note that when
        ``slice_time_ref`` and ``condition_tr_shift`` are 0, the formula simplifies to:

        ::

            start_scan = math.floor(onset / tr)
            end_scan = math.ceil((onset + duration) / tr)
            scans.extend(range(onset_scan, end_scan))
            scans = sorted(list(set(scans)))

        Filtering a specific condition from the timeseries is done after nuisance regression. The
        indices corresponding to the condition are used to extract the TRs (the timepoints that
        fall within the the event window(s) adjusted by the slice timing reference
        (``slice_time_ref``) and a fixed hemodynamic delay (``condition_tr_shift``) if specified)
        from the timeseries.

        If the "use_sample_mask" key in the ``fd_threshold`` dictionary is set to True, the
        truncated 2D timeseries is temporarily padded to ensure the correct rows corresponding to
        the condition are obtained.

        If the "interpolate" key in the ``fd_threshold`` dictionary is set to True, interpolation is
        performed using the full timeseries data (excluding dummy volumes) to replace only the
        censored (high-motion) volumes. Then, the indices corresponding to the condition are
        extracted from the timeseries, excluding any frames that do not have non-censored data at
        both edges.
        """
        if runs and not isinstance(runs, list):
            runs = [runs]

        # Update attributes
        self._task_info = {
            "task": task,
            "session": session,
            "runs": runs,
            "condition": condition,
            "condition_tr_shift": condition_tr_shift,
            "tr": tr,
            "slice_time_ref": slice_time_ref,
        }

        # Quick check condition_tr_shift and slice_time_ref to not cause issues later in the pipeline
        self._validate_get_bold_params()

        # Initialize new attributes
        self._subject_timeseries = {}
        self._qc = {}
        self._n_cores = n_cores

        layout = self._call_layout(bids_dir, pipeline_name)
        subj_ids = sorted(
            layout.get(
                return_type="id", target="subject", task=task, space=self._space, suffix="bold"
            )
        )

        if not subj_ids:
            msg = (
                "No subject IDs found - potential reasons:\n"
                "1. Incorrect template space (default: 'MNI152NLin2009cAsym'). "
                "Fix: Set correct template space using "
                "`self.space = 'TEMPLATE_SPACE'` (e.g. 'MNI152NLin6Asym')\n"
                "2. File names do not contain specific entities required for "
                "querying such as 'sub-', 'space-', 'task-', or 'desc-' "
                "(e.g 'sub-01_ses-1_task-rest_space-MNI152NLin2009cAsym_desc-preproc-bold.nii.gz')\n"
                "3. Incorrect task name specified in `task` parameter.\n"
                "4. The cache may need to be cleared using "
                "``TimeseriesExtractor._call_layout.cache_clear()`` if the directory has been "
                "changed (e.g. new files added, file names changed, etc) during the current Python "
                "session."
            )
            raise BIDSQueryError(msg)

        if exclude_subjects:
            exclude_subjects = (
                exclude_subjects if isinstance(exclude_subjects, list) else [exclude_subjects]
            )
            exclude_subjects = [
                subj_id.removeprefix("sub-") for subj_id in map(str, exclude_subjects)
            ]
            subj_ids = sorted([subj_id for subj_id in subj_ids if subj_id not in exclude_subjects])

        if run_subjects:
            run_subjects = run_subjects if isinstance(run_subjects, list) else [run_subjects]
            run_subjects = [subj_id.removeprefix("sub-") for subj_id in map(str, run_subjects)]
            subj_ids = sorted([subj_id for subj_id in subj_ids if subj_id in run_subjects])

        # Setup extraction
        self._subject_ids, self._subject_info = bids_query.setup_extraction(
            layout,
            subj_ids,
            self._space,
            exclude_niftis,
            self._signal_clean_info,
            self._task_info,
            verbose,
        )

        if self._n_cores:
            # Generate list of tuples for each subject
            args_list = [
                (
                    subj_id,
                    self._subject_info[subj_id]["prepped_files"],
                    self._subject_info[subj_id]["run_list"],
                    self._parcel_approach,
                    self._signal_clean_info,
                    self._task_info,
                    self._subject_info[subj_id]["tr"],
                    verbose,
                    flush,
                    parallel_log_config,
                )
                for subj_id in self._subject_ids
            ]

            parallel = Parallel(return_as="generator", n_jobs=self._n_cores, backend="loky")
            outputs = tqdm(
                parallel(delayed(process_subject_runs)(*args) for args in args_list),
                desc="Processing Subjects",
                total=len(args_list),
                disable=not progress_bar,
            )

            for output in outputs:
                subject_timeseries, qc = output
                self._expand_dicts(subject_timeseries, qc)
        else:
            if parallel_log_config:
                LG.warning(
                    "`parallel_log_config` is only used for parallel processing. "
                    "The default logger can be modified by configuring either the root logger or a "
                    "logger for a specific module prior to package import."
                )

            for subj_id in tqdm(
                self._subject_ids, desc="Processing Subjects", disable=not progress_bar
            ):
                outputs = process_subject_runs(
                    subj_id=subj_id,
                    **self._subject_info[subj_id],
                    parcel_approach=self._parcel_approach,
                    signal_clean_info=self._signal_clean_info,
                    task_info=self._task_info,
                    verbose=verbose,
                    flush=flush,
                )

                self._expand_dicts(*outputs)

        return self

    def _validate_get_bold_params(self) -> None:
        """Validates ``condition_tr_shift`` and ``slice_time_ref``."""
        if self._task_info["condition_tr_shift"] != 0:
            if (
                not isinstance(self._task_info["condition_tr_shift"], int)
                or self._task_info["condition_tr_shift"] < 0
            ):
                raise ValueError(
                    "`condition_tr_shift` must be a integer value equal to or greater than 0."
                )

            if self._task_info["condition"] is None:
                LG.warning("`condition_tr_shift` specified but `condition` is None.")

        if self._task_info["slice_time_ref"] != 0:
            if not isinstance(self._task_info["slice_time_ref"], (float, int)) or (
                self._task_info["slice_time_ref"] < 0 or (self._task_info["slice_time_ref"] > 1)
            ):
                raise ValueError("`slice_time_ref` must be a numerical value from 0 to 1.")

            if self._task_info["condition"] is None:
                LG.warning("`slice_time_ref` specified but `condition` is None.")

    @staticmethod
    @lru_cache(maxsize=4)
    def _call_layout(bids_dir: str, pipeline_name: Union[str, None]) -> Any:
        """
        Returns ``BIDSLayout``.

        Raises
        ------
        ModuleNotFoundError
            The pybids package is not installed.

        Note
        ----
        The type hint for the returned (``BIDSLayout``) is not added to allow certain functions in
        ``TimeseriesExtractor`` to be used on Windows machines that do not have pybids installed.
        """
        try:
            import bids
        except ModuleNotFoundError:
            raise ModuleNotFoundError(
                "This function relies on the pybids package to query subject-specific files. "
                "If on Windows, pybids does not install by default to avoid long path error issues "
                "during installation. Try using `pip install pybids` or `pip install neurocaps[windows]`."
            )

        bids_dir = os.path.normpath(bids_dir).rstrip(os.path.sep)
        bids_dir = bids_dir.removesuffix("derivatives").rstrip(os.sep)

        if pipeline_name:
            pipeline_name = os.path.normpath(pipeline_name).lstrip(os.path.sep).rstrip(os.path.sep)
            pipeline_name = pipeline_name.removeprefix("derivatives").lstrip(os.path.sep)

            layout = bids.BIDSLayout(
                bids_dir, derivatives=os.path.join(bids_dir, "derivatives", pipeline_name)
            )
        else:
            layout = bids.BIDSLayout(bids_dir, derivatives=True)

        LG.info(f"{layout}")

        return layout

    def _expand_dicts(
        self,
        subject_timeseries: Union[SubjectTimeseries, None],
        qc: Union[dict[str, dict[str, dict[str, Union[float, int]]]], None],
    ) -> None:
        """
        Aggregates individual subject timeseries and qc dictionaries into ``self._subject_timeseries``
        and ``self._qc``.
        """
        if isinstance(subject_timeseries, dict):
            self._subject_timeseries.update(subject_timeseries)
            if qc:
                self._qc.update(qc)

    @staticmethod
    def _raise_error(attr_name: str, msg: str) -> None:
        """Raises error if ``_subject_timeseries`` pr ``self._qc`` is not available."""
        if attr_name == "_subject_timeseries":
            raise AttributeError(
                f"{msg} since `self.subject_timeseries` is None, either run `self.get_bold()` or "
                "assign a valid timeseries dictionary to `self.subject_timeseries`."
            )
        else:
            raise AttributeError(f"{msg} since `self.qc` is None, run `self.get_bold()` first.")

    def timeseries_to_pickle(self, output_dir: str, filename: Optional[str] = None) -> Self:
        """
        Save the Extracted Subject Timeseries.

        Saves the extracted timeseries stored in the ``self.subject_timeseries`` dictionary as a
        pickle file (ending in ".pkl") with Joblib.

        Parameters
        ----------
        output_dir: :obj:`str`
            Directory to save a pickle file to. The directory will be created if it does not exist.

        filename: :obj:`str` or :obj:`None`, default=None
            Name of the file. If None, will use "subject_timeseries.pkl" as default.

        Returns
        -------
        self
        """
        save_filename = self._create_output_filename(
            output_dir, filename, attr_name="_subject_timeseries", call="timeseries_to_pickle"
        )

        io_utils.serialize(self._subject_timeseries, output_dir, save_filename, use_joblib=True)

        return self

    def report_qc(
        self,
        output_dir: Optional[str] = None,
        filename: Optional[str] = None,
        return_df: bool = True,
    ) -> Union[DataFrame, None]:
        """
        Report Quality Control Information.

        Converts per-subject, per-run quality control metrics (related to framewise displacement and
        dummy volumes) stored in ``self.qc`` dictionary to a pandas DataFrame for analysis and
        reporting.

        .. note:: In versions >= 0.27.0, "NaN" is used to denote a value is not applicable.

        Parameters
        ----------
        output_dir: :obj:`str`, default=None
            Directory to save a pickle file to. The directory will be created if it does not exist.
            The file will not be saved if an output directory is not provided.

        filename: :obj:`str` or :obj:`None`, default=None
            Name of the file with or without the "csv" extension.

        return_df: :obj:`bool`, default=True
            If True, returns the dataframe.

        Returns
        -------
        pandas.Dataframe
            Pandas dataframe containing the colums: "Subject_ID", "Run", "Frames_Scrubbed",
            "Frames_Interpolated", "Mean_High_Motion_Length", "Std_High_Motion_Length", "Mean_FD",
            "Std_FD", and "N_Dummy_Scans".

        Important
        ---------
        **Reporting**: Statistics for each subject's run are specific to the specified
        ``fd_threshold`` and ``dummy_scans``. Additionally, all reported statistics for framewise
        displacement are specific to steady-state volumes (if ``dummy_scans`` is not None), as they
        are only computed after accounting for dummy volumes.

        **Note:** If ``condition`` was specified in ``self.get_bold()``, all the reported statistics
        related to framewise displacement are specific to the frames assigned to the condition (as
        determined by the equation in "Extraction of Task Conditions" in the documentation of
        ``self.get_bold()``), which are treated as one continuous event window for computational
        simplicity. The number of reported dummy volumes is not adjusted for the condition.

        **Mean & Standard Deviation of Framewise Displacement:** "Mean_FD" and "Std_FD" represent
        the statistics prior to scrubbing or interpolating.

        **Censored & Interpolated Frames**: The metrics for scrubbed frames and interpolated frames
        are independent. For instance, 2 scrubbed frames and 3 interpolated frames means that 3
        frames were interpolated while 2 were scrubbed due to excessive motion.

        **Note:** Only censored frames with bounded by non-censored frames on both sides are
        interpolated, while censored frames at the edge of the timeseries (including frames that
        border censored edges) are always scrubbed and counted in "Frames_Scrubbed". For instance,
        if the full sample mask is computed as ``[0, 0, 1, 0, 0, 1, 0]`` where "0" are censored and
        "1" is not censored, then when no interpolation is requested, then "Frames_Scrubbed" will be
        5 and "Frames_Interpolated" will be 0. If interpolation is requested, then "Frames_Scrubbed"
        would be 3 and "Frames_Interpolated" would be 2 (indexes 3 and 4).

        **High Motion Length Computation**: "Mean_High_Motion_Length" and "Std_High_Motion_Length"
        represent the average length and population standard deviation of consecutive frames flagged
        for high-motion frames, respectively.
        """
        save_filename = self._create_output_filename(
            output_dir, filename, attr_name="_qc", call="report_qc"
        )
        assert self._qc, "No quality control information to report."

        # Build df
        df = DataFrame(
            columns=[
                "Subject_ID",
                "Run",
                "Mean_FD",
                "Std_FD",
                "Frames_Scrubbed",
                "Frames_Interpolated",
                "Mean_High_Motion_Length",
                "Std_High_Motion_Length",
                "N_Dummy_Scans",
            ]
        )
        for subject in self._qc:
            for run in self._qc[subject]:
                df.loc[len(df)] = [
                    subject,
                    run,
                    self._qc[subject][run]["mean_fd"],
                    self._qc[subject][run]["std_fd"],
                    self._qc[subject][run]["frames_scrubbed"],
                    self._qc[subject][run]["frames_interpolated"],
                    self._qc[subject][run]["mean_high_motion_length"],
                    self._qc[subject][run]["std_high_motion_length"],
                    self._qc[subject][run]["n_dummy_scans"],
                ]

        if output_dir:
            df.to_csv(os.path.join(output_dir, save_filename), sep=",", index=0)

        if return_df:
            return df

    def _create_output_filename(
        self, output_dir: Union[str, None], filename: Union[str, None], attr_name: str, call: str
    ) -> Union[str, None]:
        """Checks if the required attribute is present then creates the output filename."""
        if not hasattr(self, attr_name):
            self._raise_error(
                attr_name,
                msg=(
                    "Cannot save pickle file"
                    if attr_name == "_subject_timeseries"
                    else "Cannot save csv file"
                ),
            )

        if not output_dir:
            return None
        else:
            io_utils.makedir(output_dir)

        ext = "pkl" if call == "timeseries_to_pickle" else "csv"

        if filename is None:
            save_filename = (
                "subject_timeseries.pkl" if call == "timeseries_to_pickle" else "report_qc.csv"
            )
        else:
            save_filename = f"{os.path.splitext(filename.rstrip())[0].rstrip()}.{ext}"

        return save_filename

    def visualize_bold(
        self,
        subj_id: Union[int, str],
        run: Optional[Union[int, str]] = None,
        roi_indx: Optional[Union[int, str, list[str], list[int]]] = None,
        region: Optional[str] = None,
        show_figs: bool = True,
        output_dir: Optional[str] = None,
        filename: Optional[str] = None,
        as_pickle: bool = False,
        **kwargs,
    ) -> Self:
        """
        Plot the Extracted Subject Timeseries.

        Visualize the extracted BOLD timeseries data of nodes (Region-of-Interests [ROIs]) or
        regions (anatomical regions/networks) for a specific subject and run.

        Parameters
        ----------
        subj_id: :obj:`str` or :obj:`int`
            The ID of the subject.

        run: :obj:`int`, :obj:`str`, or :obj:`None`, default=None
            The run ID of the subject to plot. Must be specified if multiple runs exist for a given
            subject.

        roi_indx: :obj:`int`, :obj:`str`, :obj:`list[int]`, :obj:`list[int]` or :obj:`None`, default=None
            The indices of the parcellation nodes to plot. See "nodes" in ``self.parcel_approach``
            for valid nodes.

        region: :obj:`str` or :obj:`None`, default=None
            The region of the parcellation to plot. If not None, all nodes in the specified region
            will be averaged then plotted. See "regions" in ``self.parcel_approach`` for valid region.

        show_figs: :obj:`bool`, default=True
            Display figures.

        output_dir: :obj:`str` or :obj:`None`, default=None
            Directory to save plot as png image. The directory will be created if it does not exist.
            If None, plot will not be saved.

        filename: :obj:`str` or :obj:`None`, default=None
            Name of the file without the extension.

        as_pickle: :obj:`bool`, default=False
            When ``output_dir`` is specified, plots are saved as pickle file, which can be further
            modified, instead of png images.

            .. versionadded:: 0.26.5

        **kwargs:
            Keyword arguments used when saving figures. Valid keywords include:

            - dpi: :obj:`int`, default=300 -- Dots per inch for the figure.
            - figsize: :obj:`tuple`, default=(11, 5) -- Size of the figure in inches.
            - bbox_inches: :obj:`str` or :obj:`None`, default="tight" -- Alters size of the
              whitespace in the saved image.

        Returns
        -------
        self

        Note
        ----
        **Parcellation Approach**: the "nodes" and "regions" subkeys are required in
        ``parcel_approach``.
        """

        if not hasattr(self, "_subject_timeseries"):
            self._raise_error(attr_name="_subject_timeseries", msg="Cannot plot bold data")

        subj_id = str(subj_id).removeprefix("sub-")
        if subj_id not in self._subject_timeseries:
            raise KeyError(f"Subject {subj_id} is not available in `self._subject_timeseries`.")

        if len(subject_runs := list(self._subject_timeseries[subj_id])) > 1 and not run:
            raise ValueError(
                "`run` must be specified when multiple runs exist. "
                f"Runs available for sub-{subj_id}: {', '.join(subject_runs)}."
            )

        if roi_indx is None and region is None:
            raise ValueError("either `roi_indx` or `region` must be specified.")

        if roi_indx is not None and region is not None:
            raise ValueError("`roi_indx` and `region` can not be used simultaneously.")

        io_utils.issue_file_warning("filename", filename, output_dir)

        # Plotting defaults
        plot_dict = resolve_kwargs(PlotDefaults.visualize_bold(), **kwargs)

        # Obtain the column indices associated with the rois
        plot_indxs = vizualization.get_plot_indxs(self._parcel_approach, roi_indx, region)

        run_name = None if not run else str(run)
        run_name = subject_runs[0] if not run_name else f"run-{run_name.removeprefix('run-')}"
        timeseries = self._subject_timeseries[subj_id][run_name]
        fig = vizualization.create_bold_figure(
            timeseries, self._parcel_approach, plot_dict["figsize"], plot_indxs, roi_indx, region
        )

        vizualization.save_bold_figure(
            fig, subj_id, run_name, output_dir, filename, plot_dict, as_pickle
        )

        PlotFuncs.show(show_figs)

        return self



"""Internal module containing functions for creating radar plots."""

import os, sys
from typing import Any, Union

import numpy as np
import pandas as pd
import plotly.express as px
import plotly.graph_objects as go
import plotly.offline as pyo
from numpy.typing import NDArray

from neurocaps.typing import ParcelApproach
from neurocaps.utils import _io as io_utils
from neurocaps.utils._logging import setup_logger
from neurocaps.utils._parcellation_validation import extract_custom_region_indices, get_parc_name

LG = setup_logger(__name__)


def update_radar_dict(
    cap_dict: dict[str, NDArray],
    radar_dict: dict,
    parcel_approach: ParcelApproach,
) -> dict:
    """
    Updates a dictionary containing information about the cosine similarity between each region
    specified in a parcellation and the positive and negative activations in a CAP vector for
    all CAPs.
    """
    for cap_name in cap_dict:
        cap_vector = cap_dict[cap_name]
        radar_dict[cap_name] = {"High Amplitude": [], "Low Amplitude": []}

        for region in radar_dict["Regions"]:
            region_mask = create_region_mask_1d(parcel_approach, region, cap_vector)

            # Get high and low amplitudes
            high_amp_vector = np.where(cap_vector > 0, cap_vector, 0)
            # Invert vector for low_amp so that cosine similarity is positive
            low_amp_vector = np.where(cap_vector < 0, -cap_vector, 0)

            # Get cosine similarity between the high amplitude and low amplitude vectors
            high_amp_cosine = compute_cosine_similarity(high_amp_vector, region_mask)
            low_amp_cosine = compute_cosine_similarity(low_amp_vector, region_mask)

            radar_dict[cap_name]["High Amplitude"].append(high_amp_cosine)
            radar_dict[cap_name]["Low Amplitude"].append(low_amp_cosine)

    return radar_dict


def create_region_mask_1d(
    parcel_approach: ParcelApproach, region: str, cap_vector: NDArray[np.floating]
) -> NDArray[np.bool_]:
    """
    Creates a 1D binary mask where 1 denotes the indices in ``cap_vector`` belonging to a
    specific network/region and "0" otherwise.
    """
    # Get the index values of nodes in each network/region
    if get_parc_name(parcel_approach) == "Custom":
        indxs = extract_custom_region_indices(parcel_approach, region)
    else:
        indxs = np.array(
            [
                value
                for value, node in enumerate(
                    parcel_approach[get_parc_name(parcel_approach)]["nodes"]
                )
                if region in node
            ]
        )

    # Create mask and set ROIs not in regions to zero and ROIs in regions to 1
    region_mask = np.zeros_like(cap_vector)
    region_mask[indxs] = 1

    return region_mask


def compute_cosine_similarity(
    amp_vector: NDArray[np.floating], region_mask: NDArray[np.bool_]
) -> np.floating:
    """Compute the cosine similarity between a vector and 1D binary mask."""
    dot_product = np.dot(amp_vector, region_mask)
    norm_region_mask = np.linalg.norm(region_mask)
    norm_amp_vector = np.linalg.norm(amp_vector)
    cosine_similarity = dot_product / (norm_amp_vector * norm_region_mask)

    return cosine_similarity


def generate_radar_plot(
    use_scatterpolar: bool,
    radar_dict: dict,
    cap_name: str,
    group_name: str,
    suffix_title: Union[str, None],
    plot_dict: dict[str, Any],
) -> go.Figure:
    """Generates radar plots."""
    if use_scatterpolar:
        # Create dataframe
        df = pd.DataFrame({"Regions": radar_dict["Regions"]})
        df = pd.concat([df, pd.DataFrame(radar_dict[cap_name])], axis=1)
        regions = df["Regions"].values

        # Initialize figure
        fig = go.Figure(layout=go.Layout(width=plot_dict["width"], height=plot_dict["height"]))

        for i in ["High Amplitude", "Low Amplitude"]:
            values = df[i].values
            fig.add_trace(
                go.Scatterpolar(
                    r=list(values),
                    theta=regions,
                    connectgaps=plot_dict["connectgaps"],
                    name=i,
                    opacity=plot_dict["opacity"],
                    marker=dict(
                        color=plot_dict["color_discrete_map"][i], size=plot_dict["scattersize"]
                    ),
                    line=dict(
                        color=plot_dict["color_discrete_map"][i], width=plot_dict["linewidth"]
                    ),
                )
            )
    else:
        n = len(radar_dict["Regions"])
        df = pd.DataFrame(
            {
                "Regions": radar_dict["Regions"] * 2,
                "Amp": radar_dict[cap_name]["High Amplitude"]
                + radar_dict[cap_name]["Low Amplitude"],
            }
        )
        df["Groups"] = ["High Amplitude"] * n + ["Low Amplitude"] * n

        fig = px.line_polar(
            df,
            r=df["Amp"].values,
            theta="Regions",
            line_close=plot_dict["line_close"],
            color=df["Groups"].values,
            width=plot_dict["width"],
            height=plot_dict["height"],
            category_orders={"Regions": df["Regions"]},
            color_discrete_map=plot_dict["color_discrete_map"],
        )

    if use_scatterpolar:
        fig.update_traces(fill=plot_dict["fill"], mode=plot_dict["mode"])
    else:
        fig.update_traces(
            fill=plot_dict["fill"],
            mode=plot_dict["mode"],
            marker=dict(size=plot_dict["scattersize"]),
        )

    # Set max value
    if "tickvals" not in plot_dict["radialaxis"] and "range" not in plot_dict["radialaxis"]:
        if use_scatterpolar:
            max_value = max(df[["High Amplitude", "Low Amplitude"]].max())
        else:
            max_value = df["Amp"].max()

        default_ticks = [max_value / 4, max_value / 2, 3 * max_value / 4, max_value]
        plot_dict["radialaxis"]["tickvals"] = [round(x, 2) for x in default_ticks]

    title_text = (
        f"{group_name} {cap_name} {suffix_title}" if suffix_title else f"{group_name} {cap_name}"
    )

    # Add additional customization
    fig.update_layout(
        title=dict(text=title_text, font=plot_dict["title_font"]),
        title_x=plot_dict["title_x"],
        title_y=plot_dict["title_y"],
        showlegend=bool(plot_dict["legend"]),
        legend=plot_dict["legend"],
        legend_title_text="Cosine Similarity",
        polar=dict(
            bgcolor=plot_dict["bgcolor"],
            radialaxis=plot_dict["radialaxis"],
            angularaxis=plot_dict["angularaxis"],
        ),
    )

    return fig


def show_radar_plot(fig: go.Figure, show_figure: bool) -> None:
    """
    Show a plotly image. Method of visualization depends on whether the current Python session is
    interactive.
    """
    if show_figure:
        if bool(getattr(sys, "ps1", sys.flags.interactive)):
            fig.show()
        else:
            pyo.plot(fig, auto_open=True)


def save_radar_plot(
    fig: go.Figure,
    output_dir: Union[str, None],
    group_name: str,
    cap_name: str,
    suffix_filename: str,
    as_html: bool,
    as_json: bool,
    scale: int,
    engine: str,
) -> None:
    """
    Saves a plotly image. Images are saves as png files by default. If ``as_html`` and ``as_json``
    are True, then ``as_html`` takes precedence and images are only saved as html files.
    """
    if output_dir:
        if as_html and as_html:
            LG.warning(
                "`as_html` and `as_json` are True. Figures will only be saved as html files."
            )

        filename = io_utils.filename(
            f"{group_name.replace(' ', '_')}_{cap_name}_radar", suffix_filename, "suffix", "png"
        )
        if as_html:
            fig.write_html(os.path.join(output_dir, filename.replace(".png", ".html")))
        elif as_json:
            fig.write_json(os.path.join(output_dir, filename.replace(".png", ".json")))
        else:
            fig.write_image(
                os.path.join(output_dir, filename),
                scale=scale,
                engine=engine,
            )
